from typing import Iterable, Any, Union, Tuple


# haven't yet figured this one out...
# from typing_extensions import Protocol
# class Expression(Protocol):
#     def __add__(self, other: Expression) -> Expression: ...
#     def __sub__(self, other: Expression) -> Expression: ...
#     def __neg__(self) -> Expression: ...
#     def __pow__(self, other: Any): Expression: ...
#     def __truediv__(self, other: Any) -> 'Matrix': ...


Expression = Any


class Symbol:
    def __init__(self, name: str) -> None: ...
    def __add__(self, other: Any) -> Expression: ...
    def __sub__(self, other: Any) -> Expression: ...
    def __neg__(self) -> Expression: ...
    def __mul__(self, other: Any) -> Expression: ...
    def __truediv__(self, other: Any) -> Expression: ...
    def __pow__(self, other: Any): Expression: ...


class Matrix:
    T: 'Matrix'
    is_Matrix: bool
    free_symbols: Iterable[Symbol]
    shape: Tuple[int, ...]

    def __init__(self, elements: Iterable) -> None: ...
    def __iter__(self) -> Expression: ...
    def __next__(self) -> Expression: ...
    def __setitem__(self, *args) -> None: ...
    def __getitem__(self, *args) -> Expression: ...
    def __matmul__(self, other: 'Matrix') -> 'Matrix': ...
    def __add__(self, other: Any) -> 'Matrix': ...
    def __sub__(self, other: Any) -> 'Matrix': ...
    def __mul__(self, other: Any) -> 'Matrix': ...
    def __truediv__(self, other: Any) -> 'Matrix': ...
    def __neg__(self) -> 'Matrix': ...
    def jacobian(self, *args, **kwargs) -> 'Matrix': ...
    def reshape(self, *args) -> 'Matrix': ...
    def dot(self, other: 'Matrix') -> Expression: ...
    def xreplace(self, *args: Any) -> 'Matrix': ...


MutableDenseMatrix = Matrix


def sin(x: Any) -> Expression: ...
def cos(x: Any) -> Expression: ...
def tan(x: Any) -> Expression: ...
def asin(x: Any) -> Expression: ...
def acos(x: Any) -> Expression: ...
def atan(x: Any) -> Expression: ...
def sqrt(x: Any) -> Expression: ...


def zeros(*x: int) -> Matrix: ...
def symbols(x: str) -> Iterable[Symbol]: ...
def diag(*x: Expression) -> Matrix: ...
def count_ops(x: Matrix) -> int: ...
def hessian(x: Expression, y: Matrix) -> Matrix: ...
def trigsimp(x: Union[Expression, Matrix]) -> Union[Expression, Matrix]: ...
def init_printing(x: bool = True) -> Any: ...
def diff(e: Expression, q: Matrix) -> Matrix: ...



"""
    Usage is along the lines of:
    ```
    >>> from optim_lib.collocation import radau_3
    >>> m.interp_q  = Constraint(m.fe, m.cp, m.vars, rule=radau_3(m.q,  m.dq))
    >>> m.interp_dq = Constraint(m.fe, m.cp, m.vars, rule=radau_3(m.dq, m.ddq))
    ```
    Or,
    ```
    >>> col = collocation.get_collocation_func('radau_3')
    ```
    Or,
    ```
    >>> collocation.check_collocation_method('radua_56')  # raises ValueError
    ```
"""
from pyomo.environ import ConcreteModel, Var, Constraint
from typing import Dict, Callable, Tuple

# eg: input is a string like 'implicit_euler'
#     output is a tuple of:
#         a function which takes two pyomo Var's and returns a function
#         an integer of the corresponding number of collocation points
_collocation_mapping: Dict[str,
                           Tuple[Callable[[Var, Var], Callable], int]] = {}


def get_ncp(collocation: str):
    check_collocation_method(collocation)
    return _collocation_mapping[collocation][1]


def get_collocation_func(collocation: str):
    check_collocation_method(collocation)
    return _collocation_mapping[collocation][0]


def check_collocation_method(collocation: str):
    if not collocation in _collocation_mapping.keys():
        raise ValueError(
            f'Invalid collocation method. Valid options are {_collocation_mapping.keys()}. Got: {collocation}'
        )


def explicit_euler(q: Var, dq: Var):
    def func(m: ConcreteModel, fe: int, cp: int, var: Var):
        assert cp == 1 and len(m.cp) == 1
        if fe > 1:
            return q[fe, cp, var] == q[fe-1, cp, var] + m.hm0 * m.hm[fe] * dq[fe-1, cp, var]
        else:
            return Constraint.Skip
    return func


_collocation_mapping['explicit_euler'] = (explicit_euler, 1)


def implicit_euler(q: Var, dq: Var):
    def func(m: ConcreteModel, fe: int, cp: int, var: Var):
        assert cp == 1 and len(m.cp) == 1
        if fe > 1:
            return q[fe, cp, var] == q[fe-1, cp, var] + m.hm0 * m.hm[fe] * dq[fe, cp, var]
        else:
            return Constraint.Skip
    return func


_collocation_mapping['implicit_euler'] = (implicit_euler, 1)


def radau_3(q: Var, dq: Var):
    R = [
        [0.19681547722366, -0.06553542585020,  0.02377097434822],
        [0.39442431473909,  0.29207341166523, -0.04154875212600],
        [0.37640306270047,  0.51248582618842,  0.11111111111111],
    ]

    def func(m: ConcreteModel, fe: int, cp: int, var: Var):
        assert 1 <= cp <= 3 and len(m.cp) == 3
        if fe > 1:
            inc = sum(R[cp-1][pp-1]*dq[fe, pp, var] for pp in m.cp)
            return q[fe, cp, var] == q[fe-1, m.cp[-1], var] + m.hm0*m.hm[fe] * inc
        else:
            return Constraint.Skip
    return func


_collocation_mapping['radau_3'] = (radau_3, 3)


def radau_2(q: Var, dq: Var):
    R = [
        [0.416666125000187, -0.083333125000187],
        [0.749999625000187,  0.250000374999812],
    ]

    def func(m: ConcreteModel, fe: int, cp: int, var: Var):
        assert 1 <= cp <= 2 and len(m.cp) == 2
        if fe > 1:
            inc = sum(R[cp-1][pp-1]*dq[fe, pp, var] for pp in m.cp)
            return q[fe, cp, var] == q[fe-1, m.cp[-1], var] + m.hm0*m.hm[fe] * inc
        else:
            return Constraint.Skip
    return func


_collocation_mapping['radau_2'] = (radau_2, 2)



from typing import Any, Iterable, Callable, Optional, Tuple, List, TypeVar, Union
import sympy as sp
import numpy as np
from pyomo.environ import ConcreteModel, Var, Set, Param, RangeSet, Constraint
from sympy import Matrix as Mat
from .visual import info, debug, warn
sp.init_printing()

# derivatives ################################################################


def deriv(expr, q: Mat, dq: Mat):
    """Take the time derivative of an expression `expr` with respect to time,
    handling the chain rule semi-correctly"""
    return (sp.diff(expr, q).T * dq)[0]  # type: ignore


def full_deriv(var, q: Mat, dq: Mat, ddq: Mat):
    return deriv(var, q, dq) + deriv(var, dq, ddq)


# rotations ##################################################################
SymOrFloat = Union[sp.Symbol, float]


def rot(θ: SymOrFloat) -> Mat:
    return sp.Matrix([
        [sp.cos(θ), sp.sin(θ)],
        [-sp.sin(θ), sp.cos(θ)],
    ])


def rot_x(θ: SymOrFloat) -> Mat:
    return sp.Matrix([
        [1,         0,           0],
        [0,  sp.cos(θ),  sp.sin(θ)],
        [0, -sp.sin(θ),  sp.cos(θ)],
    ])


def rot_y(θ: SymOrFloat) -> Mat:
    return sp.Matrix([
        [sp.cos(θ), 0, -sp.sin(θ)],
        [0,         1,          0],
        [sp.sin(θ), 0,  sp.cos(θ)],
    ])


def rot_z(θ: SymOrFloat) -> Mat:
    return sp.Matrix([
        [sp.cos(θ),  sp.sin(θ), 0],
        [-sp.sin(θ), sp.cos(θ), 0],
        [0,                  0, 1],
    ])


def euler_321(phi: SymOrFloat, theta: SymOrFloat, psi: SymOrFloat) -> Mat:
    return rot_x(phi) @ rot_y(theta) @ rot_z(psi)

# utils to find equations of motion ##########################################


def skew_symmetric(Rx_I: Mat, q: Mat, dq: Mat) -> Mat:
    """Rx_I is the 3x3 rotation from body frame `x` to the inertial `I`
    `q` is a vector of symbolic variables
    `dq` is a vector of the derivates of `q`

    The return matrix can be simplified fairly quickly with `trigsimp` if
    it's made up of ~3 or fewer rotations
    """
    dRx_I = sp.zeros(3, 3)

    for i in range(3):
        dRx_I[:, i] = Rx_I[:, i].jacobian(q) * dq

    omega_Rx = Rx_I.T @ dRx_I
    return Mat([
        omega_Rx[2, 1],
        omega_Rx[0, 2],
        omega_Rx[1, 0]
    ])


def manipulator_equation(Ek: Mat, Ep: Mat, q: Mat, dq: Mat) -> Tuple[Mat, Mat, Mat]:
    """Ek and Ep are the kinetic and potential energy of the system.
    They must be passed in as sp.Matrix types (or at least have a
    `.jacobian` method defined on them)"""
    M = sp.hessian(Ek, dq)
    N = M.shape[0]

    dM = sp.zeros(N, N)
    for i in range(N):
        for j in range(N):
            dM[i, j] = Mat([M[i, j]]).jacobian(q) @ dq

    C = dM @ dq - Ek.jacobian(q).T

    G = Ep.jacobian(q).T

    return M, C, G


def calc_velocities_and_energies(
        positions: Iterable[Mat], rotations: Iterable[Mat],
        masses: Iterable[float], inertias: Iterable[Mat],
        q: Mat, dq: Mat, g: float = 9.81
) -> Tuple[Mat, Mat, List[Mat], List[Mat]]:
    """
        Calculate and return  the kinetic and potential energies of
        a system, given lists of:
        - positions of each body (in inertial)
        - rotations for each body (from body to inertial)
        - mass of each body
        - inertia of each body (given as a 3x3 matrix)
        - q (each state, such as q = [x, y, theta])
        - dq (derivative of each state, such as dq = [dx, dy, dtheta])
    """
    from functools import reduce

    dPs = [Mat(Px_I.jacobian(q) * dq) for Px_I in positions]
    ang_vels = [
        skew_symmetric(Rx_I, q, dq) for Rx_I in rotations
    ]
    # this should be sum(), but it inits the sum with the int 0, which can't
    # be added to matrices
    Ek = reduce(lambda a, b: a + b, [
        dPx_I.T * mx * dPx_I / 2. + dωx_I.T * Ix * dωx_I / 2.
        for (dPx_I, mx, dωx_I, Ix) in zip(dPs, masses, ang_vels, inertias)
    ])
    Ep = Mat([sum(m * Mat([0, 0, g]).dot(p)
                  for (m, p) in zip(masses, positions))])

    return Ek, Ep, dPs, ang_vels


def lambdify_EOM(EOM: Union[sp.Matrix, list], vars_in_EOM: List[sp.Symbol], *,
                 display_vars: bool = False, test_func: bool = True,
                 func_map: dict = {}) -> List[Callable[..., float]]:
    """ Returns a list of functions which, when called with arguments which match
    `vars_in_EOM`, will evaluate the equations of motion specified in `EOM`.
    `display_vars` specifies whether to print out `vars_in_EOM` as a sanity check.
    `test_func` specifies whether to test that the function returns a float when
    called with float arguments

        >>> lambdify_EOM(EOM, vars_in_EOM)
        [<function _lambdifygenerated(_Dummy_833, _Dummy_834, ...
         <function _lambdifygenerated(_Dummy_851, _Dummy_852, ... 
         <function _lambdifygenerated(_Dummy_923, _Dummy_924, ...]

        >>> lambdify_EOM(EOM, vars_in_EOM[:-3])  # not all vars in EOM
        AssertionError: The function didn't return a float - it's likely ...
        """
    import pyomo.environ
    import random
    import math

    if display_vars is True:
        try:
            from IPython.core.display import display
            display(vars_in_EOM)
        except:
            print(vars_in_EOM)

    func_map = {'sin': pyomo.environ.sin,
                'cos': pyomo.environ.cos,
                'pi': math.pi, **func_map}

    if isinstance(EOM, list):
        eom = sp.Matrix(EOM)
    elif EOM.is_Matrix is False:  # EOM is a sympy object, but not a Matrix
        eom = sp.Matrix([EOM])
    else:
        eom = EOM

    if not set(eom.free_symbols).issubset(set(vars_in_EOM)):
        raise ValueError('Some symbols in the eom aren\'t in `vars_in_EOM:'
                         + str(set(eom.free_symbols).difference(set(vars_in_EOM))))

    # Why pack the arguments into a vector? well... sympy lambdifies expressions
    # by writing out a function (key part!) and then calling `eval` on it. Unfortunately,
    # writing a python function with more than 255 arguments is a SyntaxError
    funcs = [sp.lambdify([vars_in_EOM], eqn, modules=[func_map])  # type: ignore
             for eqn in eom]

    # replace with set(EOM.free_symbols).difference(set(vars_in_EOM))?
    if test_func is True:
        vals = [random.random() for _ in range(len(vars_in_EOM))]
        for func in funcs:
            ret = func(vals)
            assert type(ret) == float, "The function didn't return a float - it's likely "\
                "because there are symbolic variables in the EOM "\
                "which weren't specified in `vars_in_EOM`. Got: " + \
                str(ret)

    return funcs

# simplification #################################################################


def parsimp_worker(_arg: Tuple[Any, Any], allow_recur: bool = True):
    expr, simp_func = _arg

    if expr.is_number or expr.is_Symbol:
        return expr

    args = []

    if hasattr(expr, 'args'):
        for arg in expr.args:
            if allow_recur is True and sp.count_ops(arg) > 200:
                arg = parsimp_worker((arg, simp_func), allow_recur=False)

            args.append(simp_func(arg))
    else:
        raise ValueError(f'Something is wrong here - got an object of type {type(expr)}'
                         f' with value: {expr}')

    return expr.func(*args)


def parsimp(mat: Mat, nprocs: int, f=sp.trigsimp) -> Mat:
    import multiprocessing
    with multiprocessing.Pool(processes=nprocs) as p:
        return sp.Matrix(
            p.map(parsimp_worker, [(v, f) for v in mat])
            # if something goes wrong and you need to debug, try this same function
            # but with the line below instead of the p.map line above
            # list(map(parsimp_worker, [(v, f) for v in mat]))
        ).reshape(*mat.shape)

#     import sys
#     outvals = []
#         for i, val in enumerate(p.imap_unordered(parsimp_worker,
#                                 [(v,f,i) for (i,v) in enumerate(mat)]), 1):
#             outvals.append(val)
#             if disp_progress is True:
#                 sys.stdout.write('\rSimplifying.... {0:%} done'.format(i/len(vec)))
#     return sp.Matrix(outvals).reshape(vec.shape)


T = TypeVar('T')


def flatten(ls: Iterable[Iterable[T]]) -> List[T]:
    """
    Flatten an iterable of iterables of items into a list of items
    """
    return [item for sublist in ls for item in sublist]

# interpolation ##############################################################


PyomoThing = Union[Var, Set, Param, RangeSet]


def add_to_pyomo_model(m: ConcreteModel, prefix: str, vals: Iterable[Iterable[PyomoThing]]):
    """
    Add the objects in `vals` to the pyomo model `m`, prefixed with `prefix`. Eg:

    ```python
    utils.add_to_pyomo_model(m, self.name, [
        self.pyomo_params.values(),
        self.pyomo_sets.values(),
        self.pyomo_vars.values()
    ])
    ```

    or,

    ```python
    utils.add_to_pyomo_model(m, 'shoulder', [[q, dq, ddq], [q_set]])
    ```

    where `q, dq, ddq` are pyomo `Var`'s and `q_set` is a pyomo `Set`
    """
    from itertools import chain
    for v in chain(*vals):
        newname = f'{prefix}_{v.name}'  # type: ignore
        assert not hasattr(m, newname), \
            f'The pyomo model already has an attribute with the name "{newname}"'
        setattr(m, newname, v)


def def_var(m: ConcreteModel, name: str, indexes: tuple, func: Callable, bounds: tuple = (None, None)):
    """Define a variable, equal to the result of a function. Eg:

    >>> def_var(m, 'foot_height', (m.fe,), bounds=(0, None),
                func=lambda m,fe: sum(m.foot_height[fe,:]))

    # which is equivalent to:
    >>> m.foot_height = Var(m.fe, bounds=(0, None))
    >>> @m.Constraint(m.fe)
        def foot_height_constr(m, fe):
            return m.foot_height_dummy[fe] == sum(m.foot_height[fe,:]))
    """
    # define the variable. eg:
    # >>> m.my_var_name = Var(m.fe, m.otherindex, bounds=(0, None))
    setattr(m, name, Var(*indexes, bounds=bounds))

    # define a function to set the variable. eg:
    # >>> var[m.fe, m.otherindex] == func(m, m.fe, m.otherindex)
    var = getattr(m, name)
    setterfunc = lambda *args: var.__getitem__(args[1:]) == func(*args)

    # set the variable equal to the expression. eg:
    # >>> m.my_var_name_constr = Constraint(m.fe, m.otherindex, rule=setterfunc)
    setattr(m, name + '_constr', Constraint(*indexes, rule=setterfunc))


def fix_within(lo: float, var: Var, up: float):
    """
    >>> fix_within(3., link['q'][:,:,'z'], 6.)
    """
    var.setlb(lo)
    var.setub(up)


def add_guiding_trajectory(var: Var, trajectory: Iterable[float], *, tol: float):
    """
    Guide a variable by fixing its values to be within `+-tol` of `trajectory`.

    >>> trajectory = 1 + 0.2 * np.sin(np.linspace(0, 2*np.pi, num=len(robot.m.fe)))
    >>> add_guiding_trajectory(robot['body']['q'][:,1,'z'], trajectory, tol=0.3)
    """
    warn('utils.add_guiding_trajectory has not been tested!', once=True)
    for _var, val in zip(var, trajectory):
        fix_within(val - tol, _var, val + tol)


def copy_state_init(src: Var, dst: Var):
    """
    >>> copy_state_init(robot['UFL']['q'], robot['UFR']['q'])
    """
    for idx, var in src.iteritems():
        dst[idx].value = var.value


def remove_constraint_if_exists(m: ConcreteModel, constraint: str):
    if hasattr(m, constraint):
        debug(f'Deleting existing constraint: {constraint}')
        m.del_component(constraint)

    constraint_index = constraint + '_index'
    if hasattr(m, constraint_index):
        debug(f'Deleting existing constraint index: {constraint_index}')
        m.del_component(constraint_index)


def maybe_set_var(var: Var, value: float, skip_if_fixed: bool, skip_if_not_None: bool, fix: bool) -> None:
    """
    >>> set_var(link['q'][fe,cp,'x'], 1.0, skip_if_fixed=True, skip_if_not_None=True, fix=False)
    """
    if var.fixed is True and skip_if_fixed is True:
        return
    elif var.value is not None and skip_if_not_None is True:
        return
    else:
        var.value = value
        if fix is True:
            var.fixed = True


def constrain_total_time(m: ConcreteModel, total_time: float):
    """
    >>> constrain_total_time(robot.m, total_time = (nfe-1)*robot.m.hm0.value)
    """
    remove_constraint_if_exists(m, 'total_time_constr')
    m.total_time_constr = Constraint(
        expr=sum(m.hm[fe] for fe in m.fe if fe != 1)*m.hm0 == total_time)

# other utils for pyomo ######################################################


IPOPT_PATH: Union[str, None] = None


def set_ipopt_path(ipopt_path: str):
    global IPOPT_PATH
    IPOPT_PATH = ipopt_path


def default_solver(*,
                   max_mins: int,
                   ipopt_path: Optional[str] = None,
                   solver: str = 'ma86',
                   max_iter: int = 50_000,
                   OF_print_frequency_time: int = 10,
                   output_file: str = './.ipopt-log.txt',
                   warm_start_init_point: bool = True,
                   **kwargs):
    """
    Tested linear solvers include 'mumps', 'ma77', 'ma97', 'ma86' and 'pardiso'. ma86 seems
    to perform the best for the types of systems modelled with this library

    Some solver options include,

    OF_print_frequency_iter: pos int, default 1
        Summarizing iteration output is printed every print_frequency_iter iterations,
        if at least print_frequency_time seconds have passed since last output

    OF_print_frequency_time: pos int, default 0
        Summarizing iteration output is printed if at least print_frequency_time seconds
        have passed since last output and the iteration number is a multiple of
        print_frequency_iter

    OF_warm_start_init_point: default 'no'
        Indicates whether this optimization should use a warm start initialization, where
        values of primal and dual variables are given (e.g., from a previous optimization
        of a related problem.)

    OF_hessian_approximation: default 'exact'
        Set to 'limited-memory' for L-GBFS. Default is 'exact'

    OF_accept_every_trial_step: default 'no'
        Setting this option to "yes" essentially disables the line search and makes the
        algorithm take aggressive steps, without global convergence guarantees

    print_info_string: default 'no'
        Enables printing of additional info string at end of iteration output. This string
        contains some insider information about the current iteration. For details, see
        diagnostic tags at bottom of https://coin-or.github.io/Ipopt/OUTPUT.html

    See https://coin-or.github.io/Ipopt/OPTIONS.html for more options. Note that options
    written to the options file should be prefixed with "OF_"

    Another useful page is https://coin-or.github.io/Ipopt/OUTPUT.html, which helps
    understand IPOPT's output
    """
    import os
    from datetime import datetime
    from pyomo.opt import SolverFactory
    from typing import Any

    if ipopt_path is None:
        global IPOPT_PATH
        if IPOPT_PATH is not None:
            ipopt_path = IPOPT_PATH
        # better to make this warning, or hope that IPOPT is on their path?
        else:
            warn('No path set for ipopt. Pass argument `ipopt_path`, '
                 'or call `utils.set_ipopt_path(str)`', once=True)

    opt: Any = SolverFactory('ipopt', executable=ipopt_path)
    opt.options['print_level'] = 5
    opt.options['max_cpu_time'] = max_mins * 60
    opt.options['max_iter'] = max_iter
    opt.options['Tol'] = 1e-6
    opt.options['OF_print_timing_statistics'] = 'yes'
    opt.options['halt_on_ampl_error'] = 'yes'
    opt.options['OF_print_frequency_time'] = OF_print_frequency_time
    opt.options['OF_acceptable_tol'] = 1e-3  # default: 1e-6
    opt.options['OF_warm_start_init_point'] = 'yes' if warm_start_init_point else 'no'
    opt.options['output_file'] = os.getcwd() + '/' + output_file

    opt.options['linear_solver'] = solver
    if solver == 'ma86':
        opt.options['OF_ma86_scaling'] = 'none'
    else:
        warn(f'Got solver {solver} but don\'t have any specific flags for it.')

    for key, val in kwargs.items():
        opt.options[key] = val

    tstart = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    info(f'Optimization start time: {tstart}')
    return opt


def get_vals(var: Var, idxs: Optional[Tuple] = None) -> np.ndarray:
    m = var.model()
    assert isinstance(idxs, tuple) or idxs is None

    nfe = len(m.fe)
    ncp = len(m.cp)
    arr = np.array([var[idx].value for idx in var]).astype(float)

    if idxs is None:  # assume indexed with only fe
        assert var.dim() == 1, "This variable doesn't seem to be indexed with fe exclusively"
        return arr
    else:
        idxs = tuple(len(i) for i in idxs)

        # if m.cp wasn't given as an index
        idx_dims = next(iter(var))
        if len(idx_dims) != len(idxs) + 1:
            return arr.reshape(nfe, ncp, *idxs)
        else:
            return arr.reshape(nfe, *idxs)


def get_indexes(nfe: int, ncp: int, *, one_based: bool, skipfirst: bool) -> List[Tuple[int, int]]:
    """ Get indices to index into variables, taking care of the funky first finite element

        >>> get_indexes(nfe=4, ncp=3, one_based=True)
        [(1, 3), (2, 1), (2, 2), (2, 3), (3, 1), (3, 2), (3, 3)]
        """
    offset = 1 if one_based else 0
    return [(i + offset, j + offset) for i in range(nfe) for j in range(ncp)
            if (i > 0 or j == ncp-1 or not skipfirst)]


class MarkovBinaryWalk():
    """Generate a Markov process random walk which alternates between two states:
    `a` and `b`. If in state `a`, return 1 then stay in state `a` with probability
    `a_prob`. Otherwise, go to state `b`. If in state `b`, return 0 and stay in
    state `b` with probability `b_prob`"""

    def __init__(self, a_prob: float, b_prob: float):
        self.a_prob = a_prob
        self.b_prob = b_prob
        # self.state: Union[Literal['a'],Literal['b']] = 'a'
        self.state = 'a'

    def step(self):
        import random
        if self.state == 'a':
            self.state = 'a' if random.random() < self.a_prob else 'b'
            return 1
        else:
            self.state = 'b' if random.random() < self.b_prob else 'a'
            return 0

    def walk(self, n):
        return np.array([self.step() for _ in range(n)], dtype=int)


# https://stackoverflow.com/a/22424821/1892669
def in_ipython() -> bool:
    try:
        from IPython import get_ipython
        if 'IPKernelApp' not in get_ipython().config:  # type: ignore
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def has_variable_timestep(m: ConcreteModel) -> bool:
    return m.hm.type() == Var  # .ctype



from typing import Dict, List, Tuple
import sympy as sp
from .links import Link3D
from . import utils
from pyomo.environ import Constraint, Var
from .motor import add_torque
from .foot import add_foot


def def_leg(body: Link3D, front: bool, right: bool,
            thigh_params: Dict[str, float] = {},
            calf_params: Dict[str, float] = {}
            ) -> Tuple[Link3D, Link3D]:
    """Define a leg and attach it to the front/back right/left of `body`.
        Only really makes sense when `body` is aligned along the `x`-axis"""
    # maybe flip x (or y)
    def mfx(x): return x if front else -x
    def mfy(y): return y if right else -y

    start_I = (
        body.Pb_I +
        body.Rb_I @ sp.Matrix([mfx(body.length/2), mfy(body.radius), 0])
    )
    suffix = ('F' if front else 'B') + ('R' if right else 'L')
    thigh = Link3D('U'+suffix, '-z', start_I=start_I, **thigh_params,
                   meta=['leg', 'thigh', 'front' if front else 'back', 'right' if right else 'left'])
    calf = Link3D('L'+suffix, '-z', start_I=thigh.bottom_I, **calf_params,
                  meta=['leg', 'calf', 'front' if front else 'back', 'right' if right else 'left'])
    add_foot(calf, at='bottom', nsides=8)

    body.add_hookes_joint(thigh, about='xy')
    add_torque(body, thigh, about='xy')

    thigh.add_revolute_joint(calf, about='y')
    add_torque(thigh, calf, about='y')

    return thigh, calf


def prescribe_straight_leg(upper: Var, lower: Var, fes: List[int], state: str):
    """
    >>> prescribe_straight_leg(robot['thigh']['q'], robot['calf']['q'], [4, 15], 'theta')
    which does the equivalent of,
    >>> m.straight_leg_thigh_calf = Constraint([4, 15],
    ...     rule=lambda m, fe: robot['thigh']['q'][fe,ncp,'theta'] == robot['thigh']['q'][fe,ncp,'theta'])

    or for all legs of a quadruped:

    >>> for (touchdown, liftoff), (upper, lower) in zip(foot_order_vals, (('UFL', 'LFL'), ('UFR', 'LFR'), ('UBL', 'LBL'), ('UBR', 'LBR'))):  # keep in sync with feet!
    ...     prescribe_straight_leg(robot[upper]['q'], robot[lower]['q'], [touchdown, liftoff+1])
    """
    m = upper.model()
    ncp = len(m.cp)

    name = f'straight_leg_{upper.name}_{lower.name}'

    utils.remove_constraint_if_exists(m, name)

    setattr(m, name, Constraint(fes,
                                rule=lambda m, fe: lower[fe, ncp, state] == upper[fe, ncp, state]))



"""
The *_in_air functions return the coefficient 'coeff' equal to:

    coeff = 1/2 * Cd * rho * A

It is meant to be used in the drag force equation:

    Fdrag = coeff * cos(angle) * velocity

where `angle` is the angle between the norm of the area of the object and the velocity vector

Source: https://en.wikipedia.org/wiki/Drag_equation
"""
from typing import Union
from .links import Link3D
from .system import System3D
import sympy as sp
import numpy as np
from typing import Dict, Any, Optional, List, TYPE_CHECKING
from typing_extensions import TypedDict
from . import utils
from pyomo.environ import (
    ConcreteModel, Param, Set, Var, Constraint,
)
from sympy import Matrix as Mat

if TYPE_CHECKING:
    from .variable_list import VariableList


PlotConfig = TypedDict('PlotConfig', plot_forces=bool, force_scale=float)


# From https://en.wikipedia.org/wiki/Density#Air at T = 25 C
DENSITY_OF_AIR = 1.184


def cylinder_top_in_air(A: float, *, Cd: float = 0.82, rho: float = DENSITY_OF_AIR) -> float:
    """
    Default `Cd` of 0.82 is for a "long cylinder" from the table in
    https://en.wikipedia.org/wiki/File:14ilf1l.svg#/media/File:14ilf1l.svg

    via https://en.wikipedia.org/wiki/Drag_coefficient
    """
    return 1/2 * Cd * rho * A


def cylinder_in_air(A: float, *, Cd: float = 1.1, rho: float = DENSITY_OF_AIR) -> float:
    """
    Default `Cd` of 1.1 is for a human/cables from
    https://en.wikipedia.org/wiki/Drag_coefficient#General
    """
    return 1/2 * Cd * rho * A


def _norm(vec: Mat, eps: float):
    assert vec.shape == (3, 1)
    assert eps >= 0

    x, y, z = vec
    return sp.sqrt(x**2 + y**2 + z**2 + eps)


def _angle_between(veca: Mat, vecb: Mat, eps: float):
    assert veca.shape == vecb.shape == (3, 1)

    return sp.acos(
        veca.dot(vecb) / _norm(veca, eps) / _norm(vecb, eps)
    )


class Drag3D:
    def __init__(self, name: str, r: Mat, area_norm: Mat, coeff: float, *,
                 use_dummy_vars: bool, cylinder_top: bool = False):
        """
        - name: unique name to identify the drag force
        - r: location where the drag force acts in 3D
        - area_norm: vector TODO: how to describe this, and cylinder_top?
        - coeff: all the constant stuff in a drag equation: 1/2 * Cd * rho * A

        ```
        >>> Drag3D('tail', r=[x, y, z], Rb_I=euler321(), coeff=cylinder_in_air(A))
        ```
        """
        self.name = name
        self.r = r
        self.Fmag = sp.Symbol('F_{d/%s}' % name)
        self.coeff = coeff
        self.area_norm = area_norm
        self.use_dummy_vars = use_dummy_vars
        self.cylinder_top = cylinder_top
        self.deactivated = False

        self._plot_config: PlotConfig = {
            'plot_forces': True,
            'force_scale': 1/10,
        }

    def calc_eom(self, q, dq, ddq) -> Mat:
        # angle between area and drag force (a scalar)
        # used to get the effective area
        dr_eqn = Mat(self.r.jacobian(q) * dq)

        if self.use_dummy_vars:
            from .symdef import make_xyz_syms

            self.area_norm_eqn = self.area_norm
            self.area_norm = make_xyz_syms(self.name + '-area-norm')[1]

            self.dr_eqn = dr_eqn
            self.dr = make_xyz_syms(self.name + '-dr')[1]
        else:
            self.dr = dr_eqn

        gamma = _angle_between(self.area_norm, self.dr, eps=1e-6)

        # magnitude of the drag force (a scalar) which is proportional to the velocity
        dx, dy, dz = self.dr
        if self.cylinder_top:
            utils.warn('(1 - sp.sin(gamma))**2 has not been tested properly!')
            self.Fmag_rhs = self.coeff * \
                (1 - sp.sin(gamma))**2 * (dx**2 + dy**2 + dz**2)
        else:
            self.Fmag_rhs = self.coeff * \
                sp.sin(gamma)**2 * (dx**2 + dy**2 + dz**2)

        # drag force (a vector) opposite to the velocity of the link
        self.f = - self.Fmag * self.dr / _norm(self.dr, eps=1e-6)

        # then, project the force onto the plane defined by the norm of the area
        # n = self.area_norm
        # # this assumes area_norm is a unit vector
        # self.f = self.f - n.dot(self.f) * n
        # # self.f = self.f - n.dot(self.f)/n.dot(n) * n  # this normalizes area_norm

        # input force mapping (a vector)
        self.Q = Mat((self.f.T @ self.r.jacobian(q)).T)

        return self.Q

    def add_vars_to_pyomo_model(self, m: ConcreteModel):
        Fmag = Var(m.fe, m.cp, name='Fmag', bounds=(0, None))

        xyz_set = Set(initialize=['x', 'y', 'z'] if self.use_dummy_vars else [],
                      name='xyz_set', ordered=True)
        dr = Var(m.fe, m.cp, xyz_set, name='dr')
        area_norm = Var(m.fe, m.cp, xyz_set, name='area_norm')

        self.pyomo_params: Dict[str, Param] = {}
        self.pyomo_sets: Dict[str, Set] = {'xyz_set': xyz_set}
        self.pyomo_vars: Dict[str, Var] = {
            'Fmag': Fmag,
            'dr': dr,
            'area_norm': area_norm,
        }

        utils.add_to_pyomo_model(m, self.name, [
            self.pyomo_params.values(),
            self.pyomo_sets.values(),
            self.pyomo_vars.values(),
        ])

    def get_pyomo_vars(self, fe: int, cp: int):
        """fe, cp are one-based!"""
        # NB: keep in sync with get_sympy_vars()!!
        v = self.pyomo_vars
        return [v['Fmag'][fe, cp], *v['dr'][fe, cp, :], *v['area_norm'][fe, cp, :]]

    def get_sympy_vars(self):
        # NB: keep in sync with get_pyomo_vars()!!
        return [self.Fmag] + ([*self.dr, *self.area_norm] if self.use_dummy_vars else [])

    def save_data_to_dict(self) -> Dict[str, Any]:
        m = self.pyomo_vars['Fmag'].model()
        xyz_set = self.pyomo_sets['xyz_set']
        return {
            'name': self.name,
            'Fmag': utils.get_vals(self.pyomo_vars['Fmag'], (m.cp,)),
            'dr': utils.get_vals(self.pyomo_vars['dr'], (xyz_set,)),
            'area_norm': utils.get_vals(self.pyomo_vars['area_norm'], (xyz_set,)),
            'coeff': self.coeff,
        }

    def init_from_dict_one_point(self, data: Dict[str, Any],
                                 fed: int, cpd: int,
                                 fes: Optional[int] = None, cps: Optional[int] = None,
                                 **kwargs
                                 ) -> None:
        if fes is None:
            fes = fed - 1
        if cps is None:
            cps = cpd - 1

        assert self.name == data['name']
        for attr in ('coeff',):
            if getattr(self, attr) != data[attr]:
                utils.warn(
                    f'Attribute "{attr}" of link "{self.name}" is not the same as the data: {getattr(self, attr)} != {data[attr]}')

        v = self.pyomo_vars
        utils.maybe_set_var(v['Fmag'][fed, cpd],
                            data['Fmag'][fes, cps], **kwargs)

        for axi, ax in enumerate(self.pyomo_sets['xyz_set']):
            utils.maybe_set_var(v['dr'][fed, cpd, ax],
                                data['dr'][fes, cps, axi], **kwargs)
            utils.maybe_set_var(v['area_norm'][fed, cpd, ax],
                                data['area_norm'][fes, cps, axi], **kwargs)

    def add_equations_to_pyomo_model(self,
                                     sp_variables: List[sp.Symbol],
                                     pyo_variables: 'VariableList',
                                     collocation: str):
        Fmag = self.pyomo_vars['Fmag']
        m = Fmag.model()

        if self.deactivated:
            for fe in m.fe:
                for cp in m.cp:
                    Fmag[fe, cp].fix(0)
            return

        from pyomo.environ import atan
        func_map = {
            'sqrt': lambda x: (x + 1e-6)**(1/2),
            'atan': atan,
        }
        Fmag_rhs_func = utils.lambdify_EOM(
            self.Fmag_rhs, sp_variables, func_map=func_map)[0]

        ncp = len(m.cp)

        def def_Fmag(m, fe, cp):
            if fe == 1 and cp < ncp:
                return Constraint.Skip
            else:
                return Fmag[fe, cp] == Fmag_rhs_func(pyo_variables[fe, cp])

        setattr(m, self.name + '_Fmag_constr',
                Constraint(m.fe, m.cp, rule=def_Fmag))

        if self.use_dummy_vars:
            xyz_set = self.pyomo_sets['xyz_set']
            dr = self.pyomo_vars['dr']
            dr_rhs_funcs = utils.lambdify_EOM(
                self.dr_eqn, sp_variables, func_map=func_map
            )

            def def_dr_dummy(m, fe, cp, ax):
                if fe == 1 and cp < ncp:
                    return Constraint.Skip
                else:
                    return dr[fe, cp, ax] == dr_rhs_funcs['xyz'.index(ax)](pyo_variables[fe, cp])

            setattr(m, self.name + '_dr_dummy',
                    Constraint(m.fe, m.cp, xyz_set, rule=def_dr_dummy))

            area_norm = self.pyomo_vars['area_norm']
            area_norm_rhs_funcs = utils.lambdify_EOM(
                self.area_norm_eqn, sp_variables, func_map=func_map
            )

            def def_area_norm_dummy(m, fe, cp, ax):
                if fe == 1 and cp < ncp:
                    return Constraint.Skip
                else:
                    return area_norm[fe, cp, ax] == area_norm_rhs_funcs['xyz'.index(ax)](pyo_variables[fe, cp])

            setattr(m, self.name + '_area_norm_dummy',
                    Constraint(m.fe, m.cp, xyz_set, rule=def_area_norm_dummy))

        # for animating
        self.r_func = utils.lambdify_EOM(
            self.r, sp_variables, func_map=func_map)
        self.f_func = utils.lambdify_EOM(
            self.f, sp_variables, func_map=func_map)

    def __getitem__(self, varname: str) -> Var:
        return self.pyomo_vars[varname]

    def plot_config(self, *, plot_forces: Optional[bool] = None,
                    force_scale: Optional[float] = None) -> 'Drag3D':
        if plot_forces is not None:
            self._plot_config['plot_forces'] = plot_forces

        if force_scale is not None:
            self._plot_config['force_scale'] = force_scale

        return self

    def animation_setup(self, fig, ax, data: List[List[float]]):
        if self.deactivated:
            return

        if self._plot_config['plot_forces'] is False:
            return

        self.has_line = False
        self.plot_data = np.empty((len(data), 6))
        scale = self._plot_config['force_scale']

        for fe0, d in enumerate(data):
            x, y, z = [f(d) for f in self.r_func]
            dx, dy, dz = [f(d)*scale for f in self.f_func]
            self.plot_data[fe0, :] = (x, y, z, dx, dy, dz)

    def animation_update(self, fig, ax,
                         fe: Optional[int] = None,
                         t: Optional[float] = None,
                         t_arr: Optional[np.ndarray] = None,
                         track: bool = False):
        if self.deactivated:
            return

        if self._plot_config['plot_forces'] is False:
            return

        if self.has_line:
            self.line.remove()
            self.has_line = False

        if fe is not None:
            x, y, z, dx, dy, dz = self.plot_data[fe-1]
        else:
            assert t is not None and t_arr is not None
            x, y, z, dx, dy, dz = [
                np.interp(t, t_arr, self.plot_data[:, i]) for i in range(6)]

        self.line = ax.quiver(
            x, y, z,    # <-- starting point of vector
            dx, dy, dz,  # <-- directions of vector
            arrow_length_ratio=0.05,
            color='red', alpha=.8, lw=1.5,
        )
        self.has_line = True

    def cleanup_animation(self, fig, ax):
        if self.deactivated:
            return

        try:
            del self.line
        except:
            pass

    def plot(self):
        utils.warn('Drag3D.plot() not implemented!', once=True)

    def __repr__(self) -> str:
        return f'Drag3D(name="{self.name}", coeff={self.coeff}, active={not self.deactivated})'


def add_drag(link, at: Mat, name: Optional[str] = None, **kwargs):
    if name is None:
        name = link.name + '_drag'

    assert name not in link.nodes,\
        f'This link already has a node with the name {name}'

    v = sp.zeros(3, 1)
    v['xyz'.index(link.aligned_along[1])] = 1

    if kwargs.get('cylinder_top', False):
        from math import pi
        coeff = cylinder_top_in_air(A=pi * link.radius**2)
    else:
        coeff = cylinder_in_air(A=link.length * (2 * link.radius))

    drag = Drag3D(str(name),
                  r=at,
                  area_norm=link.Rb_I * v,
                  coeff=coeff,
                  **kwargs)
    link.nodes[name] = drag
    return drag


def drag_forces(robot_or_link: Union[System3D, Link3D]) -> List[Drag3D]:
    from typing import cast
    if isinstance(robot_or_link, System3D):
        robot = robot_or_link
        return [cast(Drag3D, node)
                for link in robot.links
                for node in link.nodes.values()
                if isinstance(node, Drag3D)]
    else:
        link = robot_or_link
        return [cast(Drag3D, node)
                for node in link.nodes.values()
                if isinstance(node, Drag3D)]



# TODO: this is a work in progress!

from typing import Any


class ConcreteModel:
    def __init__(self, name: str = None) -> None: ...
    def __setattr__(self, name: str, value: Any) -> None: ...
    def __getattribute__(self, name: str) -> Any: ...
    def add_component(self, name: str, c: Constraint) -> None: ...


class Objective:
    def __init__(self, *args, **kwargs): ...


class Param:
    def __init__(self, *args, **kwargs): ...
    def model(self) -> ConcreteModel: ...


class Set:
    def __init__(self, *args, **kwargs): ...
    def __iter__(self): ...
    def __len__(self) -> int: ...


class Var:
    dim: Any
    value: float
    fixed: bool
    name: str
    def __init__(self, *args, **kwargs): ...
    def __getitem__(self, *args) -> Any: ...
    def __getindex__(self, t: tuple) -> Any: ...
    def __iter__(self): ...
    def setub(self, v: float): ...
    def setlb(self, v: float): ...
    def iteritems(self) -> Any: ...
    def model(self) -> ConcreteModel: ...


class Constraint:
    Skip: Any
    def __init__(self, *args, **kwargs): ...


class RangeSet:
    def __init__(self, *args, **kwargs): ...


class ConstraintList:
    def __init__(self, *args, **kwargs): ...


def inequality(*args: Any) -> Any: ...
def value(*args: Any) -> float: ...
def sin(*args: Any) -> Any: ...
def cos(*args: Any) -> Any: ...
def atan(*args: Any) -> Any: ...



"""
    Number of operations in EOM without simplification: 102 480. With utils.parsimp: 7776

    Example usage:
    >>> import optim_lib; utils = optim_lib.utils
    >>> robot, add_pyomo_constraints = optim_lib.models.monoped3D()
    >>> robot.calc_eom(simp_func = lambda x: utils.parsimp(x, nprocs=8))
    >>> robot.make_pyomo_model(nfe=30, collocation='euler', total_time=0.7, vary_timestep_within=(0.8, 1.2))
    >>> add_pyomo_constraints(robot)
    >>> costs = optim_lib.models.tasks.drop_test(robot, z_rot=0.2, min_torque=False, initial_height=1.0)
    >>> utils.default_solver(max_mins=10, solver='ma86', OF_hessian_approximation='limited-memory').solve(robot.m, tee=True)
    >>> robot.post_solve(costs)
    >>> robot.animate(view_along=(35, -120), t_scale=2, track='base')
"""

from typing import Tuple, Callable
from ..links import Link3D, constrain_rel_angle
from ..system import System3D
from ..foot import add_foot
from ..motor import add_torque


def model() -> Tuple[System3D, Callable[[System3D], None]]:
    # create links: body, upper leg, lower leg
    base = Link3D('base', '+x', base=True,
                  mass=5., radius=0.4, length=0.4,
                  meta=['spine'])

    upper = Link3D('upper', '-z', start_I=base.Pb_I,
                   mass=.6, radius=0.01, length=0.25,
                   meta=['leg', 'thigh'])

    lower = Link3D('lower', '-z', start_I=upper.bottom_I,
                   mass=.4, radius=0.01, length=0.25,
                   meta=['leg', 'calf'])

    add_foot(lower, at='bottom', nsides=8, friction_coeff=1.)

    # add relationships between links
    base.add_hookes_joint(upper, about='xy')
    add_torque(base, upper, about='xy',
               torque_bounds=(-2., 2.), no_load_speed=20)

    upper.add_revolute_joint(lower, about='y')
    add_torque(upper, lower, about='y', torque_bounds=(-2., 2.),
               no_load_speed=20)

    # combine into a robot
    robot = System3D('3D monoped', [base, upper, lower])

    return robot, add_pyomo_constraints


def add_pyomo_constraints(robot: System3D):
    assert robot.m is not None,\
        'robot does not have a pyomo model defined on it'

    from math import pi as π
    body, thigh, calf = [link['q'] for link in robot.links]

    constrain_rel_angle(robot.m, 'hip',
                        -π/2, body[:, :, 'theta'], thigh[:, :, 'theta'], π/2)

    constrain_rel_angle(robot.m, 'knee',
                        0, thigh[:, :, 'theta'], calf[:, :, 'theta'], π)



import sympy as sp
from sympy import Matrix as Mat
from typing import Tuple

USE_LATEX: bool = True


def make_xyz_syms(i: str) -> Tuple[Mat, Mat, Mat]:
    """
    Define x, y, z symbols and time derivatives, with subscript `i`

    ```
    >>> xyz, dxyz, ddxyz = make_xyz_syms('test')
    >>> dxyz
    Matrix([
    [\\dot{x}_{test}],
    [\\dot{y}_{test}],
    [\\dot{z}_{test}]])
    ```
    """
    global USE_LATEX
    if USE_LATEX is True:
        q = sp.symbols(r'x_{%s} y_{%s} z_{%s}' % (i, i, i))
        dq = sp.symbols(r'\dot{x}_{%s} \dot{y}_{%s} \dot{z}_{%s}' % (i, i, i))
        ddq = sp.symbols(
            r'\ddot{x}_{%s} \ddot{y}_{%s} \ddot{z}_{%s}' % (i, i, i))
    else:
        q = sp.symbols(f'  x_{i}   y_{i}   z_{i}')
        dq = sp.symbols(f' dx_{i}  dy_{i}  dz_{i}')
        ddq = sp.symbols(f'ddx_{i} ddy_{i} ddz_{i}')

    return Mat(q), Mat(dq), Mat(ddq)


def make_ang_syms(i: str) -> Tuple[Mat, Mat, Mat]:
    """
    Define `\\phi`, `\\theta`, `\\psi` (ϕ θ ψ) symbols and time derivatives, with subscript `i`

    ```
    >>> ptp, dptp, ddptp = make_ang_syms('test')
    >>> dptp
    Matrix([
    [  \\dot{\\phi}_{test}],
    [\\dot{\\theta}_{test}],
    [  \\dot{\\psi}_{test}]])
    ```
    """
    global USE_LATEX
    if USE_LATEX is True:
        q = sp.symbols(r'\phi_{%s} \theta_{%s} \psi_{%s}' % (i, i, i))
        dq = sp.symbols(
            r'\dot{\phi}_{%s} \dot{\theta}_{%s} \dot{\psi}_{%s}' % (i, i, i))
        ddq = sp.symbols(
            r'\ddot{\phi}_{%s} \ddot{\theta}_{%s} \ddot{\psi}_{%s}' % (i, i, i))
    else:
        q = sp.symbols(f'  ϕ_{i}   θ_{i}   ψ_{i}')
        dq = sp.symbols(f' dϕ_{i}  dθ_{i}  dψ_{i}')
        ddq = sp.symbols(f'ddϕ_{i} ddθ_{i} ddψ_{i}')

    return Mat(q), Mat(dq), Mat(ddq)



# this file really should be named better...
from pyomo.environ import Var
import numpy as np
from typing import Callable, Optional

from .system import System3D
from .links import Link3D
from .foot import feet, feet_penalty
from . import motor
from . import utils


def init_vel_acc_from_pos(robot: System3D, link: Link3D, varname: str):
    """
    >>> robot['LBL']['dq'][:,:,'theta'].value = 0
    >>> robot['LBL']['ddq'][:,:,'theta'].value = 0
    >>> init_vel_acc_from_pos(robot['LBL'], 'theta')
    >>> robot['LBL'].plot()  # velocities and accelerations should look right
    """
    q, dq, ddq = link['q'], link['dq'], link['ddq']
    m = q.model()
    for fe, cp in robot.indices(one_based=True):
        if fe == 1:
            continue

        dt = m.hm0.value * m.hm[fe].value

        if not dq[fe, cp, varname].fixed and q[fe, cp, varname].value is not None and q[fe-1, cp, varname].value is not None:
            dq[fe, cp, varname].value = \
                (q[fe, cp, varname].value - q[fe-1, cp, varname].value)/dt

        if not ddq[fe, cp, varname].fixed and dq[fe, cp, varname].value is not None and dq[fe-1, cp, varname].value is not None:
            ddq[fe, cp, varname].value = \
                (dq[fe, cp, varname].value - dq[fe-1, cp, varname].value)/dt


# TODO: re-add derivatives once I know their scaling is correct, and add a noise parameter!
def sin_around_touchdown(fe: int, nfe: int) -> np.ndarray:
    """Returns position, acceleration and velocity. The leg is assumed to be moving backwards
    at the time of impact with the ground, with a forward leg having a negative angle

    >>> import matplotlib.pyplot as plt
    >>> for x, label in zip(sin_around_touchdown(10, 50), ('pos', 'vel', 'acc')):
    ...     plt.plot(x, label=label)
    >>> plt.gcf().set_size_inches(25,8); plt.grid(True); plt.legend(); plt.xticks(range(50)); plt.show()
    """
    assert 1 <= fe <= nfe
    π = np.pi  # type: ignore
    offset = (fe-1)/(nfe-1) * 2*π
    x = np.linspace(0, 2*π, num=nfe) - offset
    # , np.cos(x), -np.sin(x)  <- velocity and acceleration
    return np.sin(x)  # type: ignore


def bound_penalty_and_add_transport_cost(robot: System3D, penalty_limit: float):
    from pyomo.environ import Objective
    utils.info('Deleting previous cost function')
    robot.m.del_component('cost')

    nfe = len(robot.m.fe)
    ncp = len(robot.m.cp)

    for foot in feet(robot):
        contact_penalty = foot.pyomo_vars['contact_penalty']
        friction_penalty = foot.pyomo_vars['friction_penalty']
        slip_penalty = foot.pyomo_vars['slip_penalty']

        for fe, cp in robot.indices(one_based=True):
            contact_penalty[fe].setub(penalty_limit)
            friction_penalty[fe].setub(penalty_limit)
            slip_penalty[fe, cp].setub(penalty_limit)

    body = robot.links[0]
    torque_cost = sum(
        [Tc**2 for link in robot.links for Tc in link['Tc'][:, :]])
    transport_cost = torque_cost / body['q'][nfe, ncp, 'x']
    robot.m.cost = Objective(expr=transport_cost)

    pen_cost = feet_penalty(robot)
    return {'penalty': pen_cost, 'transport': transport_cost}


def remove_contact_constraints(robot):
    for foot in feet(robot):
        fh = foot['foot_height']
        grfz = foot['GRFz']
        for fe, cp in robot.indices(one_based=True):
            fh[fe, cp].fixed = False
            fh[fe, cp].setlb(0)
            grfz[fe, cp].fixed = False
            grfz[fe, cp].setlb(0)

    for constraint in [c for c in dir(robot.m) if 'straight_leg' in c]:
        utils.remove_constraint_if_exists(robot.m, constraint)


def add_costs(robot,
              include_transport_cost: bool,
              include_torque_cost: bool,
              transport_axis: str = 'x',
              scale: float = 0.001,
              **other_costs) -> dict:  # time cost? distance cost?
    from pyomo.environ import Objective
    utils.remove_constraint_if_exists(robot.m, 'cost')

    body = robot.links[0]
    nfe = len(robot.m.fe)
    ncp = len(robot.m.cp)

    # TODO: scale by time and body weight?
#     torque_cost = sum([Tc**2 for link in robot.links for Tc in link['Tc'][:,:]])
    torque_cost = motor.torque_squared_penalty(robot)
    # the commented out code makes things slower... hmmm. Could be worth trying with just hm0?
#     torque_cost = 0
#     for link in robot.links:
#         for fe in robot.m.fe:
#             for Tc in link.pyomo_sets['Tc_set']:
#                 torque_cost += robot.m.hm[fe] * robot.m.hm0 * link['Tc'][fe,Tc]**2

    transport_cost = torque_cost / body['q'][nfe, ncp, transport_axis]

    #pen_cost = sum(link.foot.penalty_sum() for link in robot.links if link.has_foot())
    pen_cost = feet_penalty(robot)
    robot.m.cost = Objective(expr=pen_cost
                             + (transport_cost*scale if include_transport_cost else 0)
                             + (torque_cost*scale if include_torque_cost else 0)
                             + sum(v for v in other_costs.values()))

    return {'penalty': pen_cost, 'transport': transport_cost, 'torque': torque_cost, **other_costs}


def increase_motor_limits(robot: System3D, *, torque_bound: float, no_load_speed: float):
    """
    >>> robot.make_pyomo_model(nfe=10,  collocation='implicit_euler', total_time=0.3, scale_forces_by=30.)
    >>> increase_motor_limits(robot, torque_bound=5., no_load_speed=100.)
    >>> ol.motor.torques(robot)[0]['Tc'].pprint()
    """
    assert robot.m is not None,\
        'robot does not have a pyomo model defined on it'

    for motor_ in motor.torques(robot):
        for Tc in motor_['Tc'][:, :]:
            Tc.setub(+torque_bound)
            Tc.setlb(-torque_bound)

        if hasattr(motor_, 'torque_speed_limit'):
            tsp = motor_.torque_speed_limit
            tsp.torque_bounds = (-torque_bound, torque_bound)
            tsp.no_load_speed = no_load_speed


# UNTESTED
def fecp2val(nfe: int, ncp: int):
    def _(fe: int, cp: int):
        assert 1 <= fe <= nfe and 1 <= cp <= ncp, "fe and cp are one-based"
        return (fe-1 + (cp-1)/(ncp-1))*(nfe-1)
    return _


# UNTESTED


def init_with_function(var: Var, func: Callable[[float], float],
                       lowerbound: Optional[float] = None,
                       otherinds: tuple = (),
                       fixiflt0: bool = False):  # fix if less than 0
    """ func called with 0..1
    >>> init_with_function(var, lambda i: math.cos(0.2 + 2*math.pi*i)/5, lowerbound=0)
    """
    if lowerbound is None:
        lowerbound = float('-inf')

    m = var.model()
    nfe = len(m.fe)
    ncp = len(m.cp)

    for fe, cp in utils.get_indexes(nfe, ncp, one_based=True, skipfirst=False):
        val = var.__getindex__((fe, cp, *otherinds))

        i = fe/nfe + cp/ncp  # TODO: do this calculation properly
        val.value = max(lowerbound, func(i))

        if fixiflt0 is True and val.value <= 0:
            val.fixed = True



from typing import Tuple, TYPE_CHECKING
from . import utils
from pyomo.environ import Constraint

if TYPE_CHECKING:
    from .system import System3D


def periodic(robot: 'System3D', but_not: Tuple[str, ...], but_not_vel: Tuple[str, ...] = tuple()):
    """
    Make all position and velocity states in `robot` periodic, except for the
    positions (q) defined in `but_not` and the velocities (dq) in `but_not_vel`
    ```
    >>> periodic(robot, but_not=('x',))
    >>> periodic(robot, but_not=('x', 'y'), but_not_vel=('x', 'y'))
    ```
    """
    assert robot.m is not None,\
        'robot does not have a pyomo model defined on it'

    nfe, ncp = len(robot.m.fe), len(robot.m.cp)

    utils.remove_constraint_if_exists(robot.m, 'periodic_q')
    utils.remove_constraint_if_exists(robot.m, 'periodic_dq')

    # periodic positions
    qs = [
        (link['q'][1, ncp, q], link['q'][nfe, ncp, q])
        for link in robot.links
        for q in link.pyomo_sets['q_set']
        if q not in but_not
    ]
    robot.m.add_component(
        'periodic_q', Constraint(
            range(len(qs)), rule=lambda m, i: qs[i][0] == qs[i][1])
    )

    # periodic velocities
    dqs = [
        (link['dq'][1, ncp, q], link['dq'][nfe, ncp, q])
        for link in robot.links
        for q in link.pyomo_sets['q_set']
        if q not in but_not_vel
    ]
    robot.m.add_component(
        'periodic_dq', Constraint(
            range(len(dqs)), rule=lambda m, i: dqs[i][0] == dqs[i][1])
    )



# TODO: import from `typing` when not limited by python version
from typing_extensions import Protocol
from typing import Any, Dict, List, Optional, TYPE_CHECKING
from pyomo.environ import (
    ConcreteModel, Param, Set, Var,
)
from sympy import Matrix as Mat
import sympy as sp
import numpy as np

if TYPE_CHECKING:
    from .variable_list import VariableList


class Node(Protocol):
    name: str
    pyomo_params: Dict[str, Param]
    pyomo_sets: Dict[str, Set]
    pyomo_vars: Dict[str, Var]
    _plot_config: Dict[str, Any]

    def __init__(self, name: str, *args, **kwargs) -> None:
        """
        Initialize the component, setting up symbolic variables and as
        many of the equations as possible. If possible, add typing to
        the variables and `trigsimp` as you go

        Also, add `assert` statements for common errors
        """

    def calc_eom(self, q, dq, ddq) -> Mat:
        """
        Calculate the Q input mapping defined by this component,
        and return it
        """

    def add_vars_to_pyomo_model(self, m: ConcreteModel) -> None:
        """
        Define the pyomo variables that will be used in the pyomo model,
        and store them in three dictionaries

        If the component has sub components (eg a link having a foot),
        make sure to call the method on them too
        """

    def get_pyomo_vars(self, fe: int, cp: int) -> List[Var]:
        """
        Get the pyomo variables defined by this model. Note that
        `fe` and `cp` are one-based.
        NB: keep in sync with get_sympy_vars(). Eg:

        ```python
        fe -= 1
        cp -= 1

        v = self.pyomo_vars
        return [
            *v['my_var'][fe,cp,:],
        ]
        ```
        """

    def get_sympy_vars(self) -> List[sp.Symbol]:
        """
        Get the sympy variables defined by this model.
        NB: keep in sync with get_pyomo_vars()
        """

    def save_data_to_dict(self) -> Dict[str, Any]:
        """
        Save the data in this model to a dictionary, saving whatever
        is "useful" in with whatever keys you want

        NB keep in sync with `init_from_dict_one_point`
        """

    def init_from_dict_one_point(self, data: Dict[str, Any],
                                 fed: int, cpd: int,
                                 fes: Optional[int] = None, cps: Optional[int] = None,
                                 **kwargs
                                 ) -> None:
        """
        Load data, saved using the `save_data_to_dict` method

        NB keep in sync with `save_data_to_dict`
        """

    def add_equations_to_pyomo_model(self,
                                     sp_variables: List[sp.Symbol],
                                     pyo_variables: 'VariableList',
                                     collocation: str) -> None:
        """
        Add the model (and subcomponents) equations to the pyomo model.
        The sympy variables are passed so that lambda functions can be
        created. `sp_variables` will correspond to `pyo_variables`
        """

    def __getitem__(self, varname: str) -> Var:
        """
        Shorthand for `self.pyomo_vars[varname]`
        """

    def animation_setup(self, fig, ax, data: List[List[float]]):
        """
        Run setup code code an animation, like geometric calculations
        and initial setup of plot objects

        Possibly altered by options in `self._plot_config`, which is why this
        method accepts no other arguments

        TODO: explain how `data` is structured
        """

    def animation_update(self, fig, ax,
                         fe: Optional[int] = None,
                         t: Optional[float] = None,
                         t_arr: Optional[np.ndarray] = None,
                         track: bool = False):
        """
        Updates plot objects defined in `animation_setup`

        TODO: explain what the args are
        """

    def cleanup_animation(self, fig, ax):
        """
        Delete plot objects defined in `animation_setup`. Useful when jupyter
        causes handles to plot objects to remain, causing issues with python's
        garbage collector and matplotlib
        """
        try:
            del self.line
        except:
            pass

    def plot(self):
        """
        Produce line plots of the data held by the object - eg. position vs time

        Possibly altered by options in `self._plot_config`, which is why this
        method accepts no arguments
        """

    def __repr__(self) -> str:
        return f'TemplateBody(name="{self.name}")'



# It seems the discussion on when and how to add type hints to matplotlib is
# still ongoing:
# https://github.com/matplotlib/matplotlib/issues/13798
# Hopefully they'll do so soon, allowing us to delete this file

from typing import Any, Sequence, Tuple, Union


def plot(*args, **kwargs) -> Any: ...
def title(x: str) -> Any: ...
def legend(x: Union[str, Sequence[str]]) -> Any: ...
def show(x: bool = True) -> Any: ...
def grid(x: bool = True, **kwargs) -> Any: ...
def xlabel(x: str) -> Any: ...
def ylabel(x: str) -> Any: ...
def legend(*x: Any) -> Any: ...
def subplots(*args) -> Tuple[Any, Axes]: ...
def scatter(*args, **kwargs) -> Any: ...
def close(*args, **kwargs) -> Any: ...
def xlim(*args, **kwargs) -> Any: ...
def ylim(*args, **kwargs) -> Any: ...
def axis(*args, **kwargs) -> Any: ...


def figure(*args, **kwargs) -> Any: ...  # returns a Figure object
def gca(*args, **kwargs) -> Axes: ...  # returns an Axes object


class Axes:
    def plot(*args: Any, **kwargs: Any) -> Any: ...
    def scatter(*args: Any, **kwargs: Any) -> Any: ...
    def legend(*args: Any, **kwargs: Any) -> Any: ...

    def plot_surface(*args: Any, **kwrags: Any) -> Any: ...

    def view_init(self, *args, **kwargs) -> Any: ...

    def set_xlim(self, *args) -> Any: ...
    def set_ylim(self, *args) -> Any: ...
    def set_zlim(self, *args) -> Any: ...

    def get_xlim(self) -> Any: ...
    def get_ylim(self) -> Any: ...
    def get_zlim(self) -> Any: ...

    def set_xticks(self, *args) -> Any: ...
    def set_yticks(self, *args) -> Any: ...
    def set_zticks(self, *args) -> Any: ...

    def set_xlabel(self, *args) -> Any: ...
    def set_ylabel(self, *args) -> Any: ...
    def set_zlabel(self, *args) -> Any: ...

    def grid(self, x: bool = True) -> Any: ...
    def twinx(self, *args) -> Any: ...
    def _get_lines(self, *args) -> Any: ...



"""
Some tasks for the robots in the models folder.
"""
from typing import Dict, Any
from .. import utils
from pyomo.environ import Objective
from ..foot import feet_penalty
from ..motor import torque_squared_penalty


def drop_test(robot, *, z_rot: float, min_torque: bool, initial_height: float = 1.) -> Dict[str, Any]:
    """Params which have been tested for this task:
    nfe = 20, total_time = 1.0, vary_timestep_with=(0.8,1.2), 5 mins for solving

    if min_torque is True, quite a bit more time is needed as IPOPT refines things
    """
    nfe = len(robot.m.fe)
    ncp = len(robot.m.cp)

    tested_models = ('3D monoped', '3D biped',
                     '3D quadruped', '3D prismatic monoped')
    if not robot.name in tested_models:
        utils.warn('This robot configuration hasn\'t been tested!')

    body = robot['base_B'] if robot.name == '3D quadruped' else robot['base']

    # start at the origin
    body['q'][1, ncp, 'x'].fix(0)
    body['q'][1, ncp, 'y'].fix(0)
    body['q'][1, ncp, 'z'].fix(initial_height)

    # fix initial angle
    for link in robot.links:
        for ang in ('phi', 'theta'):
            link['q'][1, ncp, ang].fix(0)

        link['q'][1, ncp, 'psi'].fix(z_rot)

    # start stationary
    for link in robot.links:
        for q in link.pyomo_sets['q_set']:
            link['dq'][1, ncp, q].fix(0)

    # init to y plane
    for link in robot.links:
        for ang in ('phi', 'theta'):
            link['q'][:, :, ang].value = 0

        link['q'][:, :, 'psi'].value = z_rot

    # legs slightly forward at the end
    uplopairs = (('upper', 'lower'),) if robot.name == '3D monoped' \
        else (('UL', 'LL'), ('UR', 'LR')) if robot.name == '3D biped' \
        else (('UFL', 'LFL'), ('UFR', 'LFR'), ('UBL', 'LBL'), ('UBR', 'LBR')) if robot.name == '3D quadruped' \
        else tuple()  # <- iterating over this will result in the body not being evaluated

    for upper, lower in uplopairs:
        ang = 0.01 if not (
            robot.name == '3D quadruped' and upper[1] == 'B') else -0.01
        robot[upper]['q'][nfe, ncp, 'theta'].setlb(ang)
        robot[lower]['q'][nfe, ncp, 'theta'].setub(-ang)

    # but not properly fallen over
    body['q'][nfe, ncp, 'z'].setlb(0.2)

    # objective: reduce CoT, etc
    utils.remove_constraint_if_exists(robot.m, 'cost')

    torque_cost = torque_squared_penalty(robot)
    pen_cost = feet_penalty(robot)
    robot.m.cost = Objective(expr=(torque_cost if min_torque else 0)
                             + 1000*pen_cost)

    return {'torque': torque_cost, 'penalty': pen_cost}



from typing import Iterable, Set, Union

VALID_META: Set[str] = set((
    'spine', 'leg', 'tail',
    'front', 'back',
    'left', 'right',
    'thigh', 'calf',
))


def is_valid(meta: Union[str, Iterable[str]]) -> bool:
    if isinstance(meta, str):
        return meta in VALID_META
    else:
        return all(is_valid(m) for m in meta)



# https://stackoverflow.com/questions/15853469/putting-current-class-as-return-type-annotation
# can't use this while supporting python 3.6 (PyPy 7.3), have to use strings for now
# from __future__ import annotations
from typing import Any, Dict, Iterable, Optional, List, cast
from typing_extensions import TypedDict
import sympy as sp
import numpy as np
import pyomo.environ as pyo
from sympy import Matrix as Mat
from . import utils, symdef, visual
from .template import Node
from .variable_list import VariableList

PlotConfig = TypedDict('PlotConfig', shape=str, linewidth=float, color=str)


class Link3D:
    def __init__(self, name: str, aligned_along: str, *,
                 start_I: Optional[Mat] = None,
                 base: bool = False,
                 meta: Iterable[str] = tuple(),
                 mass: Optional[float] = None,
                 length: Optional[float] = None,
                 radius: Optional[float] = None
                 ):
        """
        Define a 3D link. Assumes the link is vertical, ie aligned with the z-axis

        name:
            name of the link
        start_I:
            the location of the start of the link in inertial coordinates
        base:
            whether this link is the/a base link
        meta:
            information which qualitatively describes a link, like 'spine', 'leg', etc
        """
        # assert sum([top_I is not None, bottom_I is not None, base is True]) == 1,\
        #     'Only one of top_I, bottom_I or base can be specified'
        assert (start_I is not None) ^ (base is True), \
            'A link must either be given an offset, or be a base link'

        acceptable_args = ('+x', '+y', '+z', '-x', '-y', '-z')
        assert aligned_along in acceptable_args,\
            f'Align the link along the x, y or z axes. Got {aligned_along}, must be one of {acceptable_args}'

        self.mass_sym = sp.Symbol('m_{%s}' % name)
        self.length_sym = sp.Symbol('L_{%s}' % name)
        self.radius_sym = sp.Symbol('r_{%s}' % name)

        self.mass, self.length, self.radius = mass, length, radius

        self.name = name
        self.is_base = base

        self.q, self.dq, self.ddq = symdef.make_ang_syms(name)
        angles = self.q  # used for defining rotation matrix

        if base is True:
            xyz, dxyz, ddxyz = symdef.make_xyz_syms(name)
            self.q = Mat([*xyz,   *self.q])
            self.dq = Mat([*dxyz,  *self.dq])
            self.ddq = Mat([*ddxyz, *self.ddq])
        else:
            xyz = None

        # rotation matrix from body to inertial
        self.Rb_I = utils.euler_321(*angles).T

        I_len = (self.mass_sym * self.radius_sym**2)/2
        I_wid = (self.mass_sym * self.length_sym**2)/12
        if aligned_along.endswith('x'):
            self.inertia = sp.diag(I_len, I_wid, I_wid)
            offset_b = Mat([-self.length_sym/2, 0, 0])
        elif aligned_along.endswith('y'):
            self.inertia = sp.diag(I_wid, I_len, I_wid)
            offset_b = Mat([0, -self.length_sym/2, 0])
        else:
            self.inertia = sp.diag(I_wid, I_wid, I_len)
            offset_b = Mat([0, 0, -self.length_sym/2])

        if aligned_along.startswith('-'):
            offset_b = -offset_b

        # positions in body and inertial axes
        if self.is_base:
            assert xyz is not None  # a type hint for mypy
            self.Pb_I = xyz
            self.top_I = self.Pb_I + self.Rb_I @ offset_b
        else:
            assert start_I is not None  # a type hint for mypy
            self.top_I = start_I
            self.Pb_I = self.top_I - self.Rb_I @ offset_b

        self.bottom_I = self.Pb_I - self.Rb_I @ offset_b

        # defining other things for later use
        self.constraint_forces: List = []
        self.angle_constraints: List = []
        self._plot_config: PlotConfig = {
            'shape': 'line',
            'linewidth': 2.,
            'color': 'tab:orange',
        }

        self.aligned_along = aligned_along

        from typing import Set as _Set
        from . import meta as _meta
        self.meta: _Set[str] = set(meta)
        assert _meta.is_valid(self.meta)

        self.nodes: Dict[str, Node] = {}

    def calc_eom(self, q, dq, ddq) -> Mat:
        if len(self.angle_constraints) > 0:
            J_c = Mat(self.angle_constraints).jacobian(q)
            Fr = Mat(self.constraint_forces)
            Q = J_c.T @ Fr
        else:
            Q = sp.zeros(len(q), 1)

        for node in self.nodes.values():
            Q += node.calc_eom(q, dq, ddq)

        return Q

    def add_hookes_joint(self, otherlink: 'Link3D', about: str) -> 'Link3D':
        """Add a hooke's joint about axis `about` of `self`
        >>> link_body.add_hookes_joint(link_UFL, about='xy')
        """
        assert all(ax in ('x', 'y', 'z') for ax in about)
        ang_constr = (self.Rb_I @ Mat([1, 0, 0])
                      ).dot(otherlink.Rb_I @ Mat([0, 1, 0]))
        self.angle_constraints.append(ang_constr)

        self.constraint_forces.append(
            sp.Symbol('F_{r/%s/%s}' % (self.name, otherlink.name))
        )

        return self

    def add_revolute_joint(self, otherlink: 'Link3D', about: str) -> 'Link3D':
        """Adds a revolute connection about axis `about` of `self`

        >>> link_UFL.add_revolute_joint(link_LFL, about='y')
        """
        assert about in ('x', 'y', 'z')
        axes = Mat([1, 0, 0]), Mat([0, 1, 0]), Mat([0, 0, 1])
        constraint_axis = axes['xyz'.index(about)]
        other_axes = [mat for idx, mat in enumerate(
            axes) if idx != 'xyz'.index(about)]

        self.angle_constraints.extend([
            (self.Rb_I @ constraint_axis).dot(otherlink.Rb_I @ other_axes[0]),
            (self.Rb_I @ constraint_axis).dot(otherlink.Rb_I @ other_axes[1]),
        ])

        self.constraint_forces.extend(
            sp.symbols('F_{r/%s/%s/:2}' % (self.name, otherlink.name))
        )

        return self

    def add_vars_to_pyomo_model(self, m: pyo.ConcreteModel) -> None:
        for attr in ('mass', 'length', 'radius'):
            assert isinstance(getattr(self, attr), float),\
                f'The {attr} for {self.name} must be set to a float'

        if self.is_base:
            q_set = pyo.Set(initialize=('x', 'y', 'z', 'phi', 'theta',
                                    'psi'), name='q_set', ordered=True)
        else:
            q_set = pyo.Set(initialize=('phi', 'theta', 'psi'),
                        name='q_set', ordered=True)

        Fr_set = pyo.Set(initialize=range(len(self.constraint_forces)),
                     name='Fr_set', ordered=True)

        q = pyo.Var(m.fe, m.cp, q_set, name='q', bounds=(-100, 100))
        dq = pyo.Var(m.fe, m.cp, q_set, name='dq', bounds=(-1000, 1000))
        # the bounds are more or less arbitrary!
        ddq = pyo.Var(m.fe, m.cp, q_set, name='ddq', bounds=(-5000, 5000))

        # constraint (reaction) forces (multiples of BW)
        Fr = pyo.Var(m.fe, m.cp, Fr_set, name='Fr', bounds=(-5, 5))

        mass = pyo.Param(initialize=self.mass, name='mass')
        length = pyo.Param(initialize=self.length, name='length')
        radius = pyo.Param(initialize=self.radius, name='radius')

        self.pyomo_params: Dict[str, pyo.Param] = {
            'mass': mass, 'length': length, 'radius': radius,
        }
        self.pyomo_sets: Dict[str, pyo.Set] = {
            'q_set': q_set, 'Fr_set': Fr_set,
        }
        self.pyomo_vars: Dict[str, pyo.Var] = {
            'q': q, 'dq': dq, 'ddq': ddq, 'Fr': Fr,
        }

        utils.add_to_pyomo_model(m, self.name, [
            self.pyomo_params.values(),
            self.pyomo_sets.values(),
            self.pyomo_vars.values(),
        ])

        if self.is_base:
            for fe in m.fe:
                for cp in m.cp:
                    q[fe, cp, 'z'].setlb(0)

        for node in self.nodes.values():
            node.add_vars_to_pyomo_model(m)

    def add_equations_to_pyomo_model(self,
                                     sp_variables: List[sp.Symbol],
                                     pyo_variables: VariableList,
                                     collocation: str):
        q = self.pyomo_vars['q']
        dq = self.pyomo_vars['dq']
        ddq = self.pyomo_vars['ddq']
        q_set = self.pyomo_sets['q_set']

        m = q.model()

        # add collocation
        from .collocation import get_collocation_func
        collocation_func = get_collocation_func(collocation)

        setattr(m, self.name + '_collocation_q',
                pyo.Constraint(m.fe, m.cp, q_set, rule=collocation_func(q,  dq)))
        setattr(m, self.name + '_collocation_dq',
                pyo.Constraint(m.fe, m.cp, q_set, rule=collocation_func(dq, ddq)))

        # while we're here, also make equations for the top and bottom of the link:
        self.top_I_func = utils.lambdify_EOM(self.top_I, sp_variables)
        self.bottom_I_func = utils.lambdify_EOM(self.bottom_I, sp_variables)

        for node in self.nodes.values():
            node.add_equations_to_pyomo_model(
                sp_variables, pyo_variables, collocation
            )

    def get_pyomo_vars(self, fe: int, cp: int) -> List:
        """fe, cp are one-based!"""
        # NB: keep in sync with get_sympy_vars()!!
        node_vars = []
        for node in self.nodes.values():
            node_vars.extend(node.get_pyomo_vars(fe, cp))

        v = self.pyomo_vars
        p = self.pyomo_params
        frset = self.pyomo_sets['Fr_set']
        return [
            *v['q'][fe, cp, :],
            *v['dq'][fe, cp, :],
            *v['ddq'][fe, cp, :],
            # using *v['Fr'][fe, cp, :] caused a bug in a newer pyomo version
            # maybe switch back at some point?
            *[v['Fr'][fe, cp, f] for f in frset],
            p['mass'], p['length'], p['radius'],
            *node_vars,
        ]

    def get_sympy_vars(self) -> List[sp.Symbol]:
        # NB: keep in sync with get_pyomo_vars()!!
        node_vars = []
        for node in self.nodes.values():
            node_vars.extend(node.get_sympy_vars())

        return [
            *self.q,
            *self.dq,
            *self.ddq,
            *self.constraint_forces,
            self.mass_sym, self.length_sym, self.radius_sym,
            *node_vars,
        ]

    def __getitem__(self, varname: str) -> pyo.Var:
        return self.pyomo_vars[varname]

    def save_data_to_dict(self) -> Dict[str, Any]:
        q_set = self.pyomo_sets['q_set']

        Fr_set = self.pyomo_sets['Fr_set']

        return {
            'name': self.name,
            'is_base': self.is_base,
            'meta': self.meta,
            'mass': self.mass,
            'length': self.length,
            'radius': self.radius,
            'q': utils.get_vals(self.pyomo_vars['q'], (q_set,)),
            'dq': utils.get_vals(self.pyomo_vars['dq'], (q_set,)),
            'ddq': utils.get_vals(self.pyomo_vars['ddq'], (q_set,)),
            'Fr': utils.get_vals(self.pyomo_vars['Fr'], (Fr_set,)) if len(Fr_set) > 0 else [],
            'nodes': [node.save_data_to_dict() for node in self.nodes.values()],
        }

    def init_from_dict_one_point(self, data: Dict[str, Any],
                                 fed: int, cpd: int,
                                 fes: Optional[int] = None, cps: Optional[int] = None,
                                 **kwargs
                                 ) -> None:
        if fes is None:
            fes = fed - 1
        if cps is None:
            cps = cpd - 1

        assert self.name == data['name'] and self.is_base == data['is_base']
        for attr in ('meta', 'mass', 'length', 'radius'):
            if getattr(self, attr) != data[attr]:
                utils.warn(
                    f'Attribute "{attr}" of link "{self.name}" is not the same as the data: {getattr(self, attr)} != {data[attr]}')

        for qstr in ('q', 'dq', 'ddq'):
            for idx, state in enumerate(self.pyomo_sets['q_set']):
                utils.maybe_set_var(
                    self.pyomo_vars[qstr][fed, cpd, state], data[qstr][fes, cps, idx], **kwargs)

        if len(data['Fr']) > 0:
            for idx, F in enumerate(self.pyomo_sets['Fr_set']):
                utils.maybe_set_var(
                    self.pyomo_vars['Fr'][fed, cpd, F], data['Fr'][fes, cps, idx], **kwargs)

        for node, nodedata in zip(self.nodes.values(), data['nodes']):
            node.init_from_dict_one_point(
                nodedata, fed, cpd, fes, cps, **kwargs)

    # Optional[Union[Literal['line'],Literal['box']]]
    def plot_config(self, *, color: Optional[str] = None,
                    shape: Optional[str] = None,
                    linewidth: Optional[float] = None) -> 'Link3D':
        """Configuration for how this link should be plotted"""
        # TODO: Still thinking about what's actually useful here. Not sure if this is a
        # wise move - I essentially am just overwriting defaults, to some extent?
        if color is not None:
            self._plot_config['color'] = color

        if shape is not None:
            assert shape in ('line', 'box')
            self._plot_config['shape'] = shape
            raise NotImplementedError('Get round to adding this!')

        if linewidth is not None:
            self._plot_config['linewidth'] = linewidth

        return self

    def animation_setup(self, fig, ax, data: List[List[float]]):
        self.line = ax.plot([], [], [],
                            linewidth=self._plot_config['linewidth'],
                            color=self._plot_config['color'])[0]

        self.plot_data = [np.empty((len(data), 3)),
                          np.empty((len(data), 3))]
        for idx, d in enumerate(data):  # xyz of top and bottom of the link
            self.plot_data[0][idx, :] = [f(d) for f in self.top_I_func]
            self.plot_data[1][idx, :] = [f(d) for f in self.bottom_I_func]

        for node in self.nodes.values():
            node.animation_setup(fig, ax, data)

    def animation_update(self, fig, ax, fe: Optional[int] = None,
                         t: Optional[float] = None, t_arr: Optional[np.ndarray] = None,
                         track: bool = False):
        if fe is not None:
            top_xyz = self.plot_data[0][fe-1]
            bottom_xyz = self.plot_data[1][fe-1]
        else:
            top_xyz = [np.interp(t, t_arr, self.plot_data[0][:, i])
                       for i in range(3)]
            bottom_xyz = [np.interp(t, t_arr, self.plot_data[1][:, i])
                          for i in range(3)]

        visual.update_3d_line(self.line, top_xyz, bottom_xyz)

        if track is True:
            lim = 1.0
            if top_xyz[2] < lim:
                top_xyz = (float(top_xyz[0]), float(top_xyz[1]), lim)
            visual.track_pt(ax, top_xyz, lim=lim)  # type: ignore

        for node in self.nodes.values():
            node.animation_update(fig, ax, fe=fe, t=t, t_arr=t_arr)

    def cleanup_animation(self, fig, ax):
        try:
            del self.line
        except:
            pass
        finally:
            for node in self.nodes.values():
                node.cleanup_animation(fig, ax)

    def plot(self) -> None:
        m = self.pyomo_vars['q'].model()

        q_set = self.pyomo_sets['q_set']
        q = utils.get_vals(self.pyomo_vars['q'], (q_set,))
        dq = utils.get_vals(self.pyomo_vars['dq'], (q_set,))
        ddq = utils.get_vals(self.pyomo_vars['ddq'], (q_set,))

        import matplotlib.pyplot as plt

        def _plt(var, title: str):
            # plot using multiple axes, as in this SO link:
            # https://stackoverflow.com/a/45925049
            # or this: https://matplotlib.org/gallery/api/two_scales.html
            if self.is_base:  # plot x,y,z separately to angles
                # typing this as 'Any' because the method accesses following give
                # false warnings otherwise...
                fig, ax1 = plt.subplots()
                ax1.plot(var[:, :3])
                ax1.legend(list(q_set)[:3])
                ax1.set_ylabel('positions [m]')
                ax1.set_xlabel('Finite element')

                ax2 = ax1.twinx()
                colors = [next(ax1._get_lines.prop_cycler)['color']
                          for _ in range(3)]
                for i, color in enumerate(colors):
                    ax2.plot(var[:, 3+i], color=color)
                ax2.legend(list(q_set)[3:])
                ax2.set_ylabel('angles [rad]')

                plt.grid(True)
                plt.title(title)
                fig.tight_layout()  # otherwise the right y-label is slightly clipped
                plt.show()
            else:
                plt.plot(var)
                plt.legend(q_set)
                plt.grid(True)
                plt.title(title)
                plt.xlabel('Finite element')
                plt.ylabel('angles [rad]')
                plt.show()

        ncp = len(m.cp)

        _plt(q[:, ncp-1, :], f'State positions in {self.name}')
        _plt(dq[:, ncp-1, :], f'State velocities in {self.name}')
        _plt(ddq[:, ncp-1, :], f'State accelerations in {self.name}')

        try:
            Fr_set = self.pyomo_sets['Fr_set']
            Fr = utils.get_vals(self.pyomo_vars['Fr'], (Fr_set,))
            plt.plot(Fr[:, ncp-1, :])
            plt.legend(Fr_set)
            plt.grid(True)
            plt.title('Constraint forces in link ' + self.name)
            plt.xlabel('Finite element')
            plt.ylabel('$F_c$ [N/body_weight]')
            plt.show()
        except StopIteration:
            pass

        for node in self.nodes.values():
            node.plot()

    def __repr__(self) -> str:
        nodes = ',\n    '.join(str(node) for node in self.nodes.values())
        nodes = f', nodes=[\n    {nodes}]' if len(nodes) > 0 else ''
        return f'Link3D(name="{self.name}", isbase={self.is_base}, mass={self.mass}kg, length={self.length}m, radius={self.radius}m{nodes})'


# class PrismaticLink3D(Link3D):
#     def __init__(self, name: str, aligned_along: str, *, bounds: Tuple[float, float], **kwargs):
#         utils.warn(
#             'Prismatic hasnt been tested yet! Remove this message if all seems okay!')

#         assert 'length' in kwargs.keys()

#         self.bounds = bounds
#         self.Δlength = sp.Symbol(f'\\Delta\\ L_{name}')
#         self.dΔlength = sp.Symbol(r'\dot{\Delta\ L_{%s}}' % name)
#         self.ddΔlength = sp.Symbol(r'\ddot{\Delta\\ L_{%s}}' % name)
#         length = self.Δlength + kwargs.pop('length')
#         Link3D.__init__(self, name, aligned_along, **kwargs, length=length)

#         self.q = sp.Matrix([*self.q,   self.Δlength])
#         self.dq = sp.Matrix([*self.dq,  self.dΔlength])
#         self.ddq = sp.Matrix([*self.ddq, self.ddΔlength])

#     def add_vars_to_pyomo_model(self, m: ConcreteModel) -> None:
#         # NOTE: copied from Link3D above, but with the addition of 'ΔL' in q_set. No other differences
#         raise NotImplementedError('Update this to keep in sync with Link3D!!')

#         if self.is_base:
#             q_set = Set(initialize=('x', 'y', 'z', 'phi', 'theta',
#                                     'psi', 'ΔL'), name='q_set', ordered=True)
#         else:
#             q_set = Set(initialize=('phi', 'theta', 'psi', 'ΔL'),
#                         name='q_set', ordered=True)

#         for fe in m.fe:
#             for cp in m.cp:
#                 q[fe, cp, 'ΔL'].setlb(self.bounds[0])
#                 q[fe, cp, 'ΔL'].setub(self.bounds[1])
#                 if self.is_base:
#                     q[fe, cp, 'z'].setlb(0)

#     # def get_pyomo_vars(self, fe: int, cp: int) -> List:
#         # return Link3D.get_pyomo_vars(self, fe, cp) + [self.pyomo_vars['ΔL'][fe,cp]]

#     # def get_sympy_vars(self) -> List[sp.Symbol]:
#     #     return Link3D.get_sympy_vars(self) + [self.Δlength, self.dΔlength, self.ddΔlength]

#     def __repr__(self) -> str:
#         return 'Prismatic' + Link3D.__repr__(self)


def constrain_rel_angle(m: pyo.ConcreteModel, constr_name: str,
                        lowerbound: float, angle1: Iterable,
                        angle2: Iterable, upperbound: float):
    # TODO: maybe switch to something like:
    # >>> diffs = [ang1-ang2 for ang1,ang2 in zip(angle1, angle2)]
    # or use this:
    # https://pyomo.readthedocs.io/en/stable/library_reference/kernel/constraint.html

    # what follows is some _horribly_ hacky code...
    # one iterable each for upper and lower bounds
    ang_1_up = iter(angle1)
    ang_1_lo = iter(angle1)
    ang_2_up = iter(angle2)
    ang_2_lo = iter(angle2)

    # _fe and _cp are unused
    # TODO: switch to pyomo.environ.inequality!
    def func(m: pyo.ConcreteModel, _fe, _cp, bound: str):
        if bound == '+':
            ang1 = next(ang_1_up)
            ang2 = next(ang_2_up)
            # make sure fe/cp match up for both links
            assert ang1.index()[:2] == ang2.index()[:2]
            return ang1 - ang2 <= upperbound
        elif bound == '-':
            ang1 = next(ang_1_lo)
            ang2 = next(ang_2_lo)
            # make sure fe/cp match up for both links
            assert ang1.index()[:2] == ang2.index()[:2]
            return lowerbound <= ang1 - ang2

    name = f'rel_angle_{constr_name}'
    setattr(m, name, pyo.Constraint(m.fe, m.cp, ('+', '-'), rule=func))



import sympy as sp
import numpy as np
from typing import Any, Dict, Iterator, List, Optional, Tuple, TYPE_CHECKING, Union
from pyomo.environ import (
    ConcreteModel, Constraint, Set, Var, Param,
)
from sympy import Matrix as Mat
from . import utils, visual
from .system import System3D
from .links import Link3D

if TYPE_CHECKING:
    from .variable_list import VariableList


class _TorqueSpeedLimit:
    def __init__(self, name: str, torque_bounds: Tuple[float, float], no_load_speed: float) -> None:
        """
        Limit the input torque to:
            $$ τ = τ_s - ω / ω_n * τ_s $$

        where
            τ = input torque
            ω = rotational velocity about the same axis
            ω_n = no load speed (x-axis intercept on graphs you may see)
            τ_s = stall torque (y-axis intercept on graphs you may see)

        From http://lancet.mit.edu/motors/motors3.html
        """
        assert torque_bounds[0] < torque_bounds[1]

        self.name = name
        self.torque_bounds = torque_bounds
        self.no_load_speed = no_load_speed

        self.relative_angle_velocities: List = []  # sympy equations
        self.axes: List[str] = []

    def add_rel_vel(self, rel_velocity, axis: str):
        self.relative_angle_velocities.append(rel_velocity)
        self.axes.append(axis)

    def add_equations_to_pyomo_model(self,
                                     sp_variables: List[sp.Symbol],
                                     pyo_variables: 'VariableList',
                                     collocation: str,
                                     Tc, Tc_set):
        m = Tc.model()
        self.Tc = Tc
        self.Tc_set = Tc_set
        self.pyo_variables = pyo_variables

        visual.warn('Not sure if the equation is correct!', once=True)
        self.rel_angle_vels_f = utils.lambdify_EOM(
            self.relative_angle_velocities, sp_variables
        )

        τ_sn, τ_sp = self.torque_bounds
        ω_n = self.no_load_speed

        def torque_speed_limit_constr(m, fe, idx, posneg):
            ω = self.rel_angle_vels_f[idx](pyo_variables[fe, 1])
            τ = Tc[fe, idx]

            # return inequality(- τ_s, τ * (1 - ω / ω_n), τ_s)
            # return inequality(- τ_s, τ * ω / ω_n, τ_s)
            if posneg == '+':
                return τ <= τ_sp * (1 - ω / ω_n)
            elif posneg == '-':
                return τ >= τ_sn * (1 + ω / ω_n)
            else:
                raise ValueError(f'posneg should be "+" or "-". Got: {posneg}')

        constraintname = self.name + '_torque_speed_limit'
        assert not hasattr(m, constraintname), \
            f'The model already has a constraint with the name {constraintname}'

        setattr(m, constraintname,
                Constraint(m.fe, Tc_set, ('+', '-'), rule=torque_speed_limit_constr))

    def plot(self, _ax=None, markersize: float = 5.):
        m = self.Tc.model()
        # ncp = len(m.cp)
        data: List[List[float]] = [
            [v.value for v in self.pyo_variables[fe, 1]]
            for fe in m.fe
        ]

        τ = utils.get_vals(self.Tc, (self.Tc_set,))
        ω = np.zeros(τ.shape)
        for idx, ω_f in enumerate(self.rel_angle_vels_f):
            for fe in m.fe:
                ω[fe-1, idx] = ω_f(data[fe-1])

        import matplotlib.pyplot as plt
        if _ax is None:
            _, ax = plt.subplots()
        else:
            ax = _ax

        # plot the torque-speed curve line
        τ_sn, τ_sp = self.torque_bounds
        ω_n = self.no_load_speed
        # plt.plot([0,  2*ω_n], [τ_sp, 0], '--', color='black')
        # plt.plot([-2*ω_n, 0], [0, τ_sn], '--', color='black')
        ax.plot([0,  2*ω_n], [τ_sp, τ_sn], '--', color='black')
        ax.plot([-2*ω_n, 0], [τ_sp, τ_sn], '--', color='black')

        # torque limit line
        ax.plot([-2*ω_n, 0], [τ_sp, τ_sp], '--', color='black')
        ax.plot([0,  2*ω_n], [τ_sn, τ_sn], '--', color='black')

        # plot torque vs speed
        for i in range(τ.shape[1]):
            ax.scatter(ω[:, i], τ[:, i], s=markersize, label=f'$\\tau_{i}$')

        if _ax is not None:
            return ax

        plt.title(f'Torque speed curve in {self.name}')
        plt.xlabel('Angular velocity $\\omega$ [rad/s]')
        plt.ylabel('Input torque $\\tau$ [Nm/body_weight]')
        plt.legend()
        plt.grid(True)
        plt.show()

        return ax

    def __repr__(self) -> str:
        return f'_TorqueSpeedLimit(torque_bounds={self.torque_bounds}, no_load_speed={self.no_load_speed}, axes={self.axes})'


class Motor3D:
    def __init__(self, name: str, Rb_I: Mat,
                 *,
                 torque_bounds: Tuple[float, float],
                 no_load_speed: Optional[float] = None):
        self.name = name
        self.Rb_I = Rb_I
        self.torque_bounds = torque_bounds

        if no_load_speed is not None:
            self.torque_speed_limit = _TorqueSpeedLimit(
                name, torque_bounds, no_load_speed)

        self.input_torques: List[sp.Symbol] = []
        self.torques_on_body = Mat([0., 0., 0.])

        # a list of tuples of [torques_on_body, rotation, about] on other bodies
        self.other_bodies: List[Tuple[Mat, Mat, str]] = []

    def add_input_torques_at(self, otherlink, about: str):
        """ Add input torques between two links, about axis `about` of `self`
        >>> link_body.add_input_torques_at(link_UFL, about='xy')
        TODO: THINK MORE ABOUT WHAT THIS REALLY MEANS! The second torque should really be rotated
        """
        assert all(axis in 'xyz' for axis in about)

        τ = Mat([
            *sp.symbols(r'\tau_{%s/%s/:%s}' % (self.name, otherlink.name, len(about)))
        ])
        self.input_torques.extend(τ)

        torques_on_other_body = Mat([0, 0, 0])

        for idx, axis in enumerate(about):
            constraint_axis = Mat([0, 0, 0])
            constraint_axis['xyz'.index(axis)] = 1
            torque = τ[idx] * self.Rb_I @ Mat(constraint_axis)
            self.torques_on_body += torque
            torques_on_other_body -= torque

        self.other_bodies.append(
            (torques_on_other_body, otherlink.Rb_I, about))

    def calc_eom(self, q, dq, ddq) -> Mat:
        ang_vel_body = utils.skew_symmetric(self.Rb_I, q, dq)
        dW = Mat([self.torques_on_body.dot(ang_vel_body)])
        Q = dW.jacobian(dq).T

        for (torques_on_body, Rb_I, about) in self.other_bodies:
            ang_vel = utils.skew_symmetric(Rb_I, q, dq)
            dW = Mat([torques_on_body.dot(ang_vel)])
            Q += dW.jacobian(dq).T

            if hasattr(self, 'torque_speed_limit'):
                for ax in about:
                    self.torque_speed_limit.add_rel_vel(
                        (ang_vel_body - ang_vel)['xyz'.index(ax)], ax,
                    )

        return Q

    def add_vars_to_pyomo_model(self, m: ConcreteModel) -> None:
        Tc_set = Set(initialize=range(len(self.input_torques)),
                     name='Tc_set', ordered=True)
        Tc = Var(m.fe, Tc_set, name='Tc', bounds=self.torque_bounds)

        self.pyomo_params: Dict[str, Param] = {}
        self.pyomo_sets: Dict[str, Set] = {'Tc_set': Tc_set}
        self.pyomo_vars: Dict[str, Var] = {'Tc': Tc}

        utils.add_to_pyomo_model(m, self.name, [
            self.pyomo_params.values(),
            self.pyomo_sets.values(),
            self.pyomo_vars.values(),
        ])

    def __getitem__(self, varname: str) -> Var:
        return self.pyomo_vars[varname]

    def get_pyomo_vars(self, fe: int, cp: int) -> List:
        return [
            *self.pyomo_vars['Tc'][fe, :],
        ]

    def get_sympy_vars(self) -> List[sp.Symbol]:
        return [
            *self.input_torques,
        ]

    def add_equations_to_pyomo_model(self,
                                     sp_variables: List[sp.Symbol],
                                     pyo_variables: 'VariableList',
                                     collocation: str):
        if hasattr(self, 'torque_speed_limit'):
            self.torque_speed_limit.add_equations_to_pyomo_model(
                sp_variables, pyo_variables, collocation,
                self.pyomo_vars['Tc'], self.pyomo_sets['Tc_set'])

    def save_data_to_dict(self) -> Dict[str, Any]:
        Tc_set = self.pyomo_sets['Tc_set']

        return {
            'name': self.name,
            'Tc': utils.get_vals(self.pyomo_vars['Tc'], (Tc_set,)) if len(Tc_set) > 0 else [],
        }

    def init_from_dict_one_point(self,
                                 data: Dict[str, Any],
                                 fed: int, cpd: int,
                                 fes: Optional[int] = None, cps: Optional[int] = None,
                                 **kwargs
                                 ) -> None:
        if fes is None:
            fes = fed - 1

        if cps is None:
            cps = cpd - 1

        # assert self.name == data['name']  # TODO: uncomment!

        if len(data['Tc']) > 0:
            for idx, T in enumerate(self.pyomo_sets['Tc_set']):
                utils.maybe_set_var(
                    self.pyomo_vars['Tc'][fed, T], data['Tc'][fes, idx], **kwargs)

    def torque_squared_cost(self):
        # TODO: add an option to scale by time and body weight?
        # previous tests slowed things down quite a bit, so I'm not sure
        # it's worth it, especially since we're not actually doing global
        # optimization in any case
        return sum([Tc**2 for Tc in self.pyomo_vars['Tc'][:, :]])

    def animation_setup(self, fig, ax, data: List[List[float]]):
        pass

    def animation_update(self, fig, ax,
                         fe: Optional[int] = None,
                         t: Optional[float] = None,
                         t_arr: Optional[np.ndarray] = None,
                         track: bool = False):
        pass

    def cleanup_animation(self, fig, ax):
        pass

    def plot(self) -> None:
        import matplotlib.pyplot as plt
        try:
            Tc_set = self.pyomo_sets['Tc_set']
            Tc = utils.get_vals(self.pyomo_vars['Tc'], (Tc_set,))
            plt.plot(Tc)
            plt.legend(Tc_set)
            plt.grid(True)
            plt.title('Input torques in link ' + self.name)
            plt.xlabel('Finite element')
            plt.ylabel('$T$ [Nm/body_weight]')
            plt.show()
        except StopIteration:
            pass

        if hasattr(self, 'torque_speed_limit'):
            self.torque_speed_limit.plot()

    def __repr__(self) -> str:
        if hasattr(self, 'torque_speed_limit'):
            tsl = f'\n{" "*6}{self.torque_speed_limit}'
        else:
            tsl = ''
        return f'Motor3D(name="{self.name}", torque_bounds={self.torque_bounds}{tsl})'


def add_torque(link, otherlink, about: str, name: Optional[str] = None, **kwargs):
    if name is None:
        name = f'{link.name}_{otherlink.name}_torque'

    assert name not in link.nodes,\
        f'This link already has a node with the name {name}'

    motor = Motor3D(str(name), link.Rb_I, **kwargs)
    link.nodes[name] = motor
    motor.add_input_torques_at(otherlink, about=about)
    return motor


def torques(robot_or_link: Union[System3D, Link3D]) -> List[Motor3D]:
    from typing import cast

    if isinstance(robot_or_link, System3D):
        robot = robot_or_link
        return [cast(Motor3D, node)
                for link in robot.links
                for node in link.nodes.values()
                if isinstance(node, Motor3D)]
    else:
        link = robot_or_link
        return [cast(Motor3D, node)
                for node in link.nodes.values()
                if isinstance(node, Motor3D)]


def torque_squared_penalty(robot: 'System3D'):
    return sum(
        torque.torque_squared_cost()
        for torque in torques(robot)
    )


# TODO: this outputs an Iterator with `len(m.fe) * len(Tc_set)` elements
# perhaps an Iterator of Iterators would make more sense?
def power(motor: Motor3D, pyo_variables, fe: Optional[int] = None) -> Iterator:
    if hasattr(motor, 'torque_speed_limit'):
        rel_angle_vels_f = motor.torque_speed_limit.rel_angle_vels_f

        Tc = motor.pyomo_vars['Tc']
        Tc_set = motor.pyomo_sets['Tc_set']
        m = Tc.model()
        v = pyo_variables

        # ω * τ
        if fe is not None:
            return (
                rel_angle_vels_f[idx](v[fe, 1]) * Tc[fe, idx]
                for idx in Tc_set
            )
        else:
            return (
                rel_angle_vels_f[idx](v[fe, 1]) * Tc[fe, idx]
                for fe in m.fe
                for idx in Tc_set
            )
    else:
        raise RuntimeError(
            'Current impementation requires a TorqueSpeedLimit :/')


def work_penalty(robot: 'System3D'):
    return sum(
        sum(power(motor, robot.pyo_variables))
        for motor in torques(robot)
    )


def work_squared_penalty(robot: 'System3D', with_time: bool):
    if with_time:
        hm = robot.m.hm
        hm0 = robot.m.hm0.value
        nfe = len(robot.m.fe)

        # enumerate returns 0-based, pyomo expects 1-based
        return sum(
            sum(hm[(fe % nfe)+1]*hm0*P**2
                for (fe, P) in enumerate(power(motor, robot.pyo_variables)))
            for motor in torques(robot)
        )
    else:
        return sum(
            sum(P**2 for P in power(motor, robot.pyo_variables))
            for motor in torques(robot)
        )


def max_power_constraint(robot: 'System3D', maxpower: float):
    def powerlimit_f(m, fe: int):
        P = sum(
            sum(power(motor, robot.pyo_variables, fe))
            for motor in torques(robot)
        )
        return P**2 <= maxpower**2

    constraintname = f'{robot.name}_maxpower'
    assert not hasattr(robot.m, constraintname), \
        f'The model already has a constraint with the name {constraintname}'

    setattr(robot.m, constraintname,
            Constraint(robot.m.fe, rule=powerlimit_f))



from types import FrameType
from typing import Optional, Tuple, List, Set, Union, Dict, Any
from contextlib import contextmanager
import os
import inspect
import numpy as np


# printing

USE_COLOURS: bool = True
DEBUG_MODE: bool = True
ONCE_SET: Set[str] = set()


class bcolors:
    # https://stackoverflow.com/questions/287871/how-to-print-colored-text-in-terminal-in-python
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    BLINK = '\33[5m'


# fix for windows (based on stackoverflow link above). Haven't tested it myself
if os.name == 'nt':
    os.system('color')


@contextmanager
def no_colour_output():
    try:
        global USE_COLOURS
        USE_COLOURS = False
        yield
    finally:
        USE_COLOURS = True


def get_useful_info(f: FrameType) -> str:
    # s is something like:
    # <code object remove_constraint_if_exists at 0x000000000aeeb100, file "/home/alex/alknemeyer-msc/python/optim_lib/utils.py", line 301>
    s = str(f.f_back.f_back.f_code)

    # the indexing removes the leading '<code object' and trailing '>'
    # the first three underscores receive 'at 0xabcd, file'
    # the final underscore handles 'line'
    # middle is splatted this way in case there are spaces in the file path
    fn_name, _, _, _, *middle, _, lineno = s[13:-1].split()

    # indexing removes the comma at the end + surrounding quotes (regardless of if their single or double)
    filepath = ''.join(middle)[1:-2].rstrip('.py').lstrip('<').rstrip('>')

    if 'optim_lib' in filepath:
        # eg: '/home/alex/alknemeyer-msc/python/optim_lib/utils'
        filepath = filepath[filepath.index('optim_lib'):].replace('/', '.')
    elif 'ipython' in filepath:
        # eg: 'ipython-input-34-ce5fcbaca0c1'
        filepath = filepath[:filepath.rfind('-')]

    return f'{filepath}.{fn_name}(), line {lineno}: '


def _printutil(prefix: str, s: str, colour: str, once: bool):
    if once is True:
        if s in ONCE_SET:
            return
        else:
            ONCE_SET.add(s)

    try:
        # Ref: https://stackoverflow.com/a/57712700/
        # calling_func = f'{inspect.currentframe().f_back.f_back.f_code}(): '
        frame = inspect.currentframe()
        if frame is not None:
            calling_func = get_useful_info(frame)
        else:
            calling_func = ''
    except:
        calling_func = ''

    if USE_COLOURS is True:
        print(bcolors.BOLD + colour + str(prefix) + bcolors.ENDC,
              colour + calling_func + str(s) + bcolors.ENDC)
    else:
        print(str(prefix), calling_func + str(s))


def info(s: str, prefix: str = '[INFO 💡]', once: bool = False):
    _printutil(prefix, s, bcolors.OKBLUE, once)


def warn(s: str, prefix: str = '[WARNING]', once: bool = False):
    _printutil(prefix, s, bcolors.FAIL, once)


def error(s: str, prefix: str = '[ERROR💩]', once: bool = False):
    _printutil(prefix, s, bcolors.FAIL + bcolors.BLINK, once)


def debug(s: str, prefix: str = '[DEBUG🐞]', once: bool = False):
    if DEBUG_MODE is True:
        _printutil(prefix, s, bcolors.WARNING, once)


def success(s: str, prefix: str = '[SUCCESS]', once: bool = False):
    _printutil(prefix, s, bcolors.OKGREEN, once)


def test_coloring():
    info('using xyz parameter...')
    warn('this might not work :/')
    error('divided by zero :o')
    debug('the value is 2.5')
    success('it worked!')

    warn('this should print once!', once=True)
    warn('this should print once!', once=True)
    warn('this should print once!', once=True)
    warn('this should print once!', once=True)


def print_data_summary(data: Dict[str, Any], indent: int = 0) -> None:
    """Print a more summarized version of the nested dictionaries and lists returned
       by calling `robot.print_data_summary()`"""
    for key, val in data.items():
        if isinstance(val, np.ndarray):
            dims = 'x'.join(map(str, val.shape))
            print(' '*indent + f'{key}: [{dims} np.ndarray]')

        elif isinstance(val, list) and len(val) > 0 and isinstance(val[0], dict):
            print(' ' * indent + f'{key}: [ ', end='')
            for v in val:
                print('{')
                print_data_summary(v, indent + 4)
                print(' '*(indent+2) + '} ', end='')
            print('\n' + ' '*indent + ']')

        elif isinstance(val, dict):
            print(' '*indent + f'{key}: ' + '{')
            print_data_summary(val, indent + 2)
            print(' '*indent + '}')

        else:
            print(' '*indent + f'{key}: {val}')

# matplotlib stuff


def plot3d_setup(figsize=(10, 10),
                 dpi: int = 60,
                 title: Optional[str] = None,
                 show_grid: bool = True,
                 xyz_labels: Optional[Tuple[str, str, str]] = None,
                 scale_plot_size: bool = True):
    """Set up a fairly "standard" figure for a 3D system, including:
    - setting up the title correctly
    - setting limits on and labelling axes
    - optionally removing the grid
    - plotting the ground (at z = 0m)
    - scaling the plot to reduce surrounding whitespace

    Returns a figure and axis
    """
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=figsize, dpi=dpi)
    ax = plt.gca(projection='3d')
    fig.tight_layout()

    if title is not None:
        fig.suptitle(title, fontsize=20)

    if xyz_labels is None:
        ax.set_xlabel('$X$ [m]')
        ax.set_ylabel('$Y$ [m]')
        ax.set_zlabel('$Z$ [m]')
    else:
        ax.set_xlabel(xyz_labels[0])
        ax.set_ylabel(xyz_labels[1])
        ax.set_zlabel(xyz_labels[2])

    if show_grid is False:
        ax.grid(False)     # hide grid lines
        ax.set_xticks([])  # hide axes ticks
        ax.set_yticks([])
        ax.set_zticks([])
        plt.axis('off')    # hmmm
        plt.grid(b=None)   # also, hmm

    def add_ground(ground_lims: Tuple[Tuple[float, float], Tuple[float, float]],
                   color: str = 'green',
                   alpha: float = 0.2):
        """ground_lims could be something like, (-10*lim, lim), (-lim, 10*lim)"""
        xx, yy = ground_lims
        ground = ax.plot_surface(*np.meshgrid(xx, yy),
                                 np.zeros((2, 2)),
                                 alpha=alpha,
                                 color=color,
                                 zorder=1)
        return ground

    if scale_plot_size is True:
        fig.subplots_adjust(left=-0.25, bottom=-0.25, right=1.25, top=1.25,
                            wspace=None, hspace=None)

    return fig, ax, add_ground


def update_3d_line(line, pt1: List[float], pt2: List[float]):
    """Update data in a 3D line, passed as two points"""
    line.set_data([[pt1[0], pt2[0]],
                   [pt1[1], pt2[1]]])
    line.set_3d_properties([pt1[2], pt2[2]])


def set_view(ax, along: Union[Tuple[float, float], str] = (45, 150)):
    """Set the angle for the 'camera' in matplotlib"""
    if type(along) == tuple:
        ax.view_init(elev=along[0], azim=along[1])

    else:
        assert along in ('x', 'y', 'z')
        if along == 'x':
            ax.view_init(elev=0, azim=0)
            ax.set_xlabel('')
        elif along == 'y':
            ax.view_init(elev=0, azim=-90)
            ax.set_ylabel('')
        elif along == 'z':
            ax.view_init(elev=90, azim=0)
            ax.set_zlabel('')


def track_pt(ax, pt: Union[Tuple[float, float], Tuple[float, float, float]], lim: float):
    """Adjust the limits of a plot so as to track a 2D or 3D point"""
    ax.set_xlim(pt[0]-lim, pt[0]+lim)
    ax.set_ylim(pt[1]-lim, pt[1]+lim)
    if len(pt) > 2:
        ax.set_zlim(pt[2]-lim, pt[2]+lim)


def get_limsx(link, dy: float = 2., z: float = 2) -> Tuple[Tuple, Tuple, Tuple]:
    """Get limits for a movement along the positive `x`-axis. Use like,
    >>> robot.animate(lims=get_lims(robot.links[0],dy=4),
                      view_along=(35, -120),
                      plot3d_config={'figsize': (15, 3), 'dpi': 100})
    """
    from .utils import get_vals
    xdata = get_vals(link['q'], (link.pyomo_sets['q_set'],))[:, 0, 0]
    xlim = max(xdata) + 1
    return ((0, xlim), (-dy/2, dy/2), (0, z))


def get_limsy(link, dx: float = 2., z: float = 2) -> Tuple[Tuple, Tuple, Tuple]:
    """Get limits for a movement along the positive `y`-axis. Use like,
    >>> robot.animate(lims=get_lims(robot.links[0],dx=4),
                      view_along=(35, -120),
                      plot3d_config={'figsize': (15, 3), 'dpi': 100})
    """
    from .utils import get_vals
    ydata = get_vals(link['q'], (link.pyomo_sets['q_set'],))[:, 0, 1]
    ylim = max(ydata) + 1
    return ((-dx/2, dx/2), (0, ylim), (0, z))


def get_limsxy(link, z: float = 2) -> Tuple[Tuple, Tuple, Tuple]:
    from .utils import get_vals
    xdata = get_vals(link['q'], (link.pyomo_sets['q_set'],))[:, 0, 0]
    xlims = (min(xdata) - 1, max(xdata) + 1)
    ydata = get_vals(link['q'], (link.pyomo_sets['q_set'],))[:, 0, 1]
    ylims = (min(ydata) - 1, max(ydata) + 1)
    return (xlims, ylims, (0, z))



"""
Running optim_lib as main currently doesn't do anything, but it could be useful
to set up as some easy way of running experiments in parallel? Eg. something like,

    $ python -m optim_lib setup_file.py --parallel-solves 4 --output-dir logs/ --until-successfull-solves 100 --save-model

which would expect a certain file structure in setup_file.py, solving 4 problems at a time until
there are 100 successfull solves, each time saving the model and IPOPT logs to logs/

setup_file.py might look like,
```
import optim_lib
# other imports...

def get_model() -> System3D:
    # define model, which could just be loading it from a pickle file
    # this gets run once
    with open('cheetah.robot', 'rb') as f:
        robot, add_pyomo_constraints = dill.load(f)
    return robot

def setup_model(robot: System3D) -> pyomo.environ.Solver:
    # make a pyomo model and add constraints
    robot.make_pyomo_model(nfe=50, collocation='euler', total_time=0.25)
    add_pyomo_constraints(robot)
    periodic_gallop_test(robot, avg_vel=10, include_cot=True)
    return solver
```

So the script would run something like,

>>> master_robot = get_model()
>>> for i in range(N):
...     robot = copy.deepcopy(master_robot)
...     solver = setup_robot(robot)
...     solver.solve(robot.m, tee=False)

except in parallel

None of that has been implemented, but it could be a useful future thing to do!
"""



"""A library for research into legged critters"""
__version__ = '0.1.1'

from . import (
    foot,
    leg,
    links,
    system,
    tasks,
    utils,
    models,
    drag,
    motor,
    visual,
    init_tools,
)

__all__ = [
    'foot',
    'leg',
    'links',
    'system',
    'tasks',
    'utils',
    'models',
    'drag',
    'motor',
    'visual',
    'init_tools'
]

# TODO: Rename library to something like rigid_body_traj_opt (import as rbt)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['physical_education', 'physical_education.models']

package_data = \
{'': ['*']}

install_requires = \
['dill ~= 0.3',
 'matplotlib ~= 3.1',
 'numpy ~= 1.10',
 'pyomo == 5.6.8',
 'sympy ~= 1.3',
 'typing_extensions ~= 3.0']

entry_points = \
{'console_scripts': ['flit = flit:main']}

setup(name='physical_education',
      version='0.1.1',
      description='A library for research into legged critters',
      author='Alexander Knemeyer',
      author_email='alknemeyer@hey.com',
      url='https://github.com/alknemeyer/physical_education',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



from typing import Any, Dict, Iterable, Optional, Tuple, Callable
from math import pi as π
from sympy import Matrix as Mat
from ..links import Link3D, constrain_rel_angle
from ..system import System3D
from ..foot import add_foot
from ..motor import add_torque
from ..drag import add_drag


parameters = {
    'model-6': {
        'source': """
            A model of cheetah 6 from Functional anatomy of the cheetah (Acinonyx jubatus) forelimb and hindlimb
            doi: 10.1111/j.1469-7580.2011.01344.x and 10.1111/j.1469-7580.2010.01310.x
        """,
        'body_B': {'mass': 17., 'radius': 0.08, 'length': 0.41},
        'body_F': {'mass': 8., 'radius': 0.08, 'length': 0.21},
        'tail0': {'mass': 0.4, 'radius': 0.005, 'length': 0.38},
        'tail1': {'mass': 0.2, 'radius': 0.005, 'length': 0.38},
        'front': {
            'thigh': {'mass': 0.171, 'radius': 0.012, 'length': 0.254},
            'calf':  {'mass': 0.068, 'radius': 0.005, 'length': 0.247},
        },
        'back': {
            'thigh': {'mass': 0.210, 'radius': 0.010, 'length': 0.281},
            'calf':  {'mass': 0.160, 'radius': 0.011, 'length': 0.287},
        },
        'friction_coeff': 1.3,
        'motor_params': {'torque_bounds': (-2., 2.), 'no_load_speed': 50.},
    },
    'mean-male': {
        'source': """
        Parameters for the 'mean' (X) cheetah from
        Morphology, Physical Condition, and Growth of the Cheetah (Acinonyx jubatus jubatus)
        https://academic.oup.com/jmammal/article/84/3/840/905900

            body mass = 45.6 kg  ---> majority (42kg?) in body
            chest girth = 71.7 cm  ---> front radius = 0.717m / (2*pi)
            abdomen girth = 59.4 cm  ---> back radius = 0.594m / (2*pi)
            skull length = 23.4 cm
            body length = 125.5 cm  ---> body - skull - neck = 125.5 - 23.4 - (20?) = 80cm => front = 0.5m, back = 0.3m
            tail length = 76.7 cm  ---> 38cm per half
            total length = 202.2 cm
            total foreleg length = 77 cm
            total hind leg length = 81.1 cm
            front foot length = 8.2 cm
            front foot width = 6.1 cm
            hind foot length = 9.2 cm
            hind foot width = 6.2 cm

        From "Quasi-steady state aerodynamics of the cheetah tail"
            fur length on tail = 10mm on average
            average tail diameter (no fur) = 31mm
                ---> radius = 31/2 + 10 = 25.5mm = 0.0255m

        Friction coeff of 1.3 from
        "Locomotion dynamics of hunting in wild cheetahs"

        NOTE: leg measurements mostly cribbed from 'model-6' above. Find proper values!
              lengths = same
              masses = same * 1.2
              radii = same

        NOTE: the motor_params values are mostly made up. In any case, different muscle
              groups would need different values
        """,
        'body_B': {'mass': 28., 'radius': 0.594/(2*π), 'length': 0.5},
        'body_F': {'mass': 14., 'radius': 0.717/(2*π), 'length': 0.3},
        'tail0': {'mass': 0.4, 'radius': 0.0255, 'length': 0.38},
        'tail1': {'mass': 0.2, 'radius': 0.0255, 'length': 0.38},
        'front': {
            'thigh': {'mass': 0.171*1.2, 'radius': 0.012, 'length': 0.254},
            'calf':  {'mass': 0.068*1.2, 'radius': 0.005, 'length': 0.247},
        },
        'back': {
            'thigh': {'mass': 0.210*1.2, 'radius': 0.010, 'length': 0.281},
            'calf':  {'mass': 0.160*1.2, 'radius': 0.011, 'length': 0.287},
        },
        'friction_coeff': 1.3,
        'motor_params': {'torque_bounds': (-3., 3.), 'no_load_speed': 100.},
    },
}


def model(params: Dict[str, Any]) -> Tuple[System3D, Callable[[System3D], None]]:
    """
    Roughly 400 000 operations in the equations of motion without simplification,
    and 142 520 if simplified with
    >>> robot.calc_eom(simp_func = lambda x: utils.parsimp(x, nprocs = 14))
    """
    # create front and back links of body and tail
    body_B = Link3D('base_B', '+x', base=True, **params['body_B'],
                    meta=['spine', 'back'])

    body_F = Link3D('base_F', '+x', start_I=body_B.bottom_I, **params['body_F'],
                    meta=['spine', 'front'])

    body_B.add_hookes_joint(body_F, about='xy')
    add_torque(body_B, body_F, about='xy', **params['motor_params'])

    tail0 = Link3D('tail0', '-x', start_I=body_B.top_I,
                   **params['tail0'], meta=['tail'])
    tail1 = Link3D('tail1', '-x', start_I=tail0.bottom_I,
                   **params['tail1'], meta=['tail'])

    # TODO: maybe rather friction = 0 ?
    add_foot(tail1, at='bottom', nsides=8, friction_coeff=0.1)

    body_B.add_hookes_joint(tail0, about='xy')
    add_torque(body_B, tail0, about='xy', **params['motor_params'])

    tail0.add_hookes_joint(tail1, about='xy')
    add_torque(tail0, tail1, about='xy', **params['motor_params'])

    add_drag(body_F, at=body_F.bottom_I, name='body_F-drag-head',
             use_dummy_vars=True, cylinder_top=True)
    add_drag(body_F, at=body_F.Pb_I,
             name='body_F-drag-body', use_dummy_vars=True)
    add_drag(body_B, at=body_B.Pb_I, use_dummy_vars=True)
    add_drag(tail0, at=tail0.Pb_I, use_dummy_vars=True)
    add_drag(tail1, at=tail1.Pb_I, use_dummy_vars=True)

    def def_leg(body: Link3D, front: bool, right: bool) -> Tuple[Link3D, Link3D]:
        """Define a leg and attach it to the front/back right/left of `body`.
            Only really makes sense when `body` is aligned along the `x`-axis"""
        # maybe flip x (or y)
        # the model is considered to face along the x axis (so front/back
        # refers to changes in the y value).
        def mfx(x): return x if front else -x
        def mfy(y): return y if right else -y

        start_I = body.Pb_I + \
            body.Rb_I @ Mat([mfx(body.length/2), mfy(body.radius), 0])
        suffix = ('F' if front else 'B') + ('R' if right else 'L')

        p = params['front'] if front else params['back']

        thigh = Link3D('U'+suffix, '-z', start_I=start_I, **p['thigh'],
                       meta=['leg', 'thigh', 'front' if front else 'back', 'right' if right else 'left'])

        calf = Link3D('L'+suffix, '-z', start_I=thigh.bottom_I, **p['calf'],
                      meta=['leg', 'calf', 'front' if front else 'back', 'right' if right else 'left'])
        add_foot(calf, at='bottom', nsides=8,
                 friction_coeff=params['friction_coeff'])

        body.add_hookes_joint(thigh, about='xy')
        add_torque(body, thigh, about='xy', **params['motor_params'])

        thigh.add_revolute_joint(calf, about='y')
        add_torque(thigh, calf, about='y', **params['motor_params'])

        return thigh, calf

    ufl, lfl = def_leg(body_F, front=True, right=False)
    ufr, lfr = def_leg(body_F, front=True, right=True)
    ubl, lbl = def_leg(body_B, front=False, right=False)
    ubr, lbr = def_leg(body_B, front=False, right=True)

    # combine into a robot
    robot = System3D('3D quadruped', [body_B, body_F, tail0, tail1,
                                      ufl, lfl, ufr, lfr,
                                      ubl, lbl, ubr, lbr])
    return robot, add_pyomo_constraints


def add_pyomo_constraints(robot: System3D) -> None:
    # π/3 = 60 degrees
    # π/2 = 90 degrees
    # π/4 = 45 degrees

    assert robot.m is not None,\
        'robot does not have a pyomo model defined on it'

    link_body_B, link_body_F, link_tail0, link_tail1, \
        link_UFL, link_LFL, link_UFR, link_LFR, \
        link_UBL, link_LBL, link_UBR, link_LBR = [
            link['q'] for link in robot.links]

    # spine can't bend too much:
    # it only has pitch and roll relative degrees of freedom. No need to constrain yaw
    constrain_rel_angle(robot.m, 'spine_pitch',
                        -π/4, link_body_B[:, :, 'theta'], link_body_F[:, :, 'theta'], π/4)
    constrain_rel_angle(robot.m, 'spine_roll',
                        -π/4, link_body_B[:, :, 'phi'], link_body_F[:, :, 'phi'], π/4)

    # tail can't go too crazy:
    constrain_rel_angle(robot.m, 'tail_body_pitch',
                        -π/3, link_body_B[:, :, 'theta'], link_tail0[:, :, 'theta'], π/3)
    constrain_rel_angle(robot.m, 'tail_body_yaw',
                        -π/3, link_body_B[:, :, 'phi'], link_tail0[:, :, 'phi'], π/3)

    constrain_rel_angle(robot.m, 'tail_tail_pitch',
                        -π/2, link_tail0[:, :, 'theta'], link_tail1[:, :, 'theta'], π/2)
    constrain_rel_angle(robot.m, 'tail_tail_yaw',
                        -π/2, link_tail0[:, :, 'phi'], link_tail1[:, :, 'phi'], π/2)

    # legs: hip abduction and knee
    for body, thigh, calf, name in ((link_body_F, link_UFL, link_LFL, 'FL'),
                                    (link_body_F, link_UFR, link_LFR, 'FR'),
                                    (link_body_B, link_UBL, link_LBL, 'BL'),
                                    (link_body_B, link_UBR, link_LBR, 'BR')):
        constrain_rel_angle(robot.m, name + '_hip_pitch',
                            -π/2, body[:, :, 'theta'], thigh[:, :, 'theta'], π/2)
        constrain_rel_angle(robot.m, name + '_hip_aduct',
                            -π/8, body[:, :, 'phi'], thigh[:, :, 'phi'], π/8)

        lo, up = (-π, 0) if name.startswith('B') else (0, π)
        constrain_rel_angle(robot.m, name + '_knee',
                            lo, thigh[:, :, 'theta'], calf[:, :, 'theta'], up)

# common functions


def high_speed_stop(robot: System3D, initial_vel: float, minimize_distance: bool,
                    gallop_data: Optional[dict] = None, offset: int = 0):
    import math
    import random
    from ..utils import copy_state_init
    from ..init_tools import add_costs

    nfe = len(robot.m.fe)
    ncp = len(robot.m.cp)
    total_time = float((nfe-1)*robot.m.hm0.value)

    body = robot['base_B']

    # start at the origin
    body['q'][1, ncp, 'x'].fix(0)
    body['q'][1, ncp, 'y'].fix(0)

    if gallop_data is not None:
        for fed, cpd in robot.indices(one_based=True):
            robot.init_from_dict_one_point(
                gallop_data, fed=fed, cpd=cpd, fes=(fed-1 + offset) % nfe, cps=0,
                skip_if_fixed=True, skip_if_not_None=False, fix=False)

        for link in robot.links:
            for q in link.pyomo_sets['q_set']:
                link['q'][1, ncp, q].fixed = True
                link['dq'][1, ncp, q].fixed = True
    else:
        # init to y plane
        body['q'][:, :, 'y'].value = 0
        for link in robot.links:
            for ang in ('phi', 'psi'):
                link['q'][:, :, ang].value = 0
                link['dq'][:, :, ang].value = 0
                link['ddq'][:, :, ang].value = 0

        # roughly bound to y plane
        for fe, cp in robot.indices(one_based=True):
            body['q'][fe, cp, 'y'].setub(0.2)
            body['q'][fe, cp, 'y'].setlb(-0.2)

        for link in robot.links:
            for ang in ('phi', 'psi'):
                for fe, cp in robot.indices(one_based=True):
                    link['q'][fe, cp, ang].setub(math.pi/4)
                    link['q'][fe, cp, ang].setlb(-math.pi/4)

        # bound theta
        for fe, cp in robot.indices(one_based=True):
            for link in robot.links[4:]:  # all leg segments - no tail or body
                link['q'][fe, cp, 'theta'].setub(math.radians(60))
                link['q'][fe, cp, 'theta'].setlb(math.radians(-60))
            for link in robot.links[:2]:  # two body segments
                link['q'][fe, cp, 'theta'].setub(math.radians(45))
                link['q'][fe, cp, 'theta'].setlb(math.radians(-45))

        for link in robot.links:
            for fe, cp in robot.indices(one_based=True):
                link['q'][fe, cp, 'theta'].value = (
                    math.radians(random.gauss(0, 15)))

        body['q'][1, ncp, 'z'].fix(1.0)
        for fe, cp in robot.indices(one_based=True):
            body['q'][fe, cp, 'z'].setub(1.2)

        # both sides mirrored
        for src, dst in (('UFL', 'UFR'), ('LFL', 'LFR'), ('UBL', 'UBR'), ('LBL', 'LBR')):
            copy_state_init(robot[src]['q'], robot[dst]['q'])

        # init tail to flick?
        for link in robot.links[2:4]:
            for fe, cp in robot.indices(one_based=True):
                link['q'][fe, cp, 'theta'].value = (
                    math.radians(random.random()*60))

    # start at speed
    body['dq'][1, ncp, 'x'].fix(initial_vel)

    # end at rest
    for link in robot.links:
        for q in link.pyomo_sets['q_set']:
            link['dq'][nfe, ncp, q].fix(0)

    # end in a fairly standard position
    for link in robot.links[:2]:  # two body segments
        link['q'][nfe, ncp, 'theta'].setub(math.radians(10))
        link['q'][nfe, ncp, 'theta'].setlb(math.radians(-10))
    for link in robot.links[4:]:  # leaving out tail - it might flail, which is good
        link['q'][nfe, ncp, 'theta'].setub(math.radians(20))
        link['q'][nfe, ncp, 'theta'].setlb(math.radians(-20))

    for link in robot.links:
        for ang in ('phi', 'psi'):
            link['q'][nfe, ncp, ang].setub(math.radians(5))
            link['q'][nfe, ncp, ang].setlb(math.radians(-5))

    # position and velocity over time
    for fe in robot.m.fe:
        pos = total_time * (initial_vel/2) * (fe-1)/(nfe-1)
        vel = initial_vel * (1 - (fe-1)/(nfe-1))
        # print('pos', pos, 'vel', vel)
        body['q'][fe, :, 'x'].value = pos
        body['dq'][fe, :, 'x'].value = vel

    # objective
    distance_cost = body['q'][nfe, ncp, 'x'] if minimize_distance else 0
    return add_costs(robot, include_transport_cost=False, include_torque_cost=False,
                     distance_cost=0.0001*distance_cost)


def periodic_gallop_test(robot,
                         avg_vel: float,
                         feet: Iterable,
                         foot_order_vals: Iterable,
                         init_from_dict: Optional[dict] = None,
                         at_angle_d: Optional[float] = None
                         ):
    """
    foot_order_vals = ((1, 7), (6, 13), (31, 38), (25, 32))  # 14 m/s
    """
    import math
    from math import radians
    import random
    from ..utils import constrain_total_time
    from ..foot import prescribe_contact_order
    from ..leg import prescribe_straight_leg
    from ..init_tools import sin_around_touchdown, add_costs
    from ..tasks import periodic

    nfe = len(robot.m.fe)
    ncp = len(robot.m.cp)
    total_time = float((nfe-1)*robot.m.hm0.value)

    constrain_total_time(robot.m, total_time=total_time)

    body = robot['base_B']

    # start at the origin
    body['q'][1, ncp, 'x'].fix(0)
    body['q'][1, ncp, 'y'].fix(0)

    if init_from_dict is None:
        if at_angle_d is None:
            # init to y plane
            body['q'][:, :, 'y'].value = 0

        # running in a straight line
        for link in robot.links:
            for ang in ('phi', 'psi'):
                link['q'][:, :, ang].value = (
                    radians(at_angle_d or 0) if ang == 'psi' else 0
                )
                link['dq'][:, :, ang].value = 0
                link['ddq'][:, :, ang].value = 0

        for fe, cp in robot.indices(one_based=True):
            var = robot.links[0]['q'][fe, cp, 'psi']
            var.setub(radians((at_angle_d or 0) + 10))
            var.setlb(radians((at_angle_d or 0) - 10))

        # init theta
        def rand(mu, sigma, offset=0):
            return radians(random.gauss(mu, sigma)+offset)

        for fe, cp in robot.indices(one_based=True):  # body
            robot.links[0]['q'][fe, cp, 'theta'].value = rand(0, 15)
            robot.links[1]['q'][fe, cp, 'theta'].value = rand(0, 15, +10)
            robot.links[2]['q'][fe, cp, 'theta'].value = rand(0, 15, -10)
            robot.links[3]['q'][fe, cp, 'theta'].value = rand(0, 15, -10)

        for link in robot.links[4:]:  # legs
            for fe, cp in robot.indices(one_based=True):
                link['q'][fe, cp, 'theta'].value = rand(0, 30)

        # body height
        body['q'][:, :, 'z'].value = 0.55

        # the feet:
        prescribe_contact_order(feet, foot_order_vals)  # type: ignore
        for (touchdown, liftoff), foot in zip(foot_order_vals, [foot.name.rstrip('_foot') for foot in feet]):
            lower, upper = foot, 'U' + foot[1:]
            prescribe_straight_leg(robot[upper]['q'], robot[lower]['q'],
                                   [touchdown], state='theta')

            angles = sin_around_touchdown(int((touchdown + liftoff)/2),
                                          len(robot.m.fe))
            for fe, val in zip(robot.m.fe, angles):
                robot[upper]['q'][fe, :, 'theta'].value = val
                robot[lower]['q'][fe, :, 'theta'].value = val + \
                    radians(-15 if upper[1] == 'F' else 15)

        # get timestep bounds ready
        # [long/short] timesteps in the air
        robot.m.hm[:].value = robot.m.hm[1].lb
        for start, stop in foot_order_vals:
            for fe in range(start, stop+1):
                # but [short/long] timesteps while on the ground
                robot.m.hm[fe].value = robot.m.hm[fe].ub
    else:
        #ol.utils.info('Using an init from a dictionary file -- assuming its an Euler init')
        for fed, cpd in robot.indices(one_based=True):
            robot.init_from_dict_one_point(init_from_dict, fed=fed, cpd=cpd, fes=fed-1,
                                           cps=0, skip_if_fixed=True, skip_if_not_None=False, fix=False)
        if not (at_angle_d == 0 or at_angle_d is None):
            raise ValueError(
                f'TODO: rotate init! Got at_angle_d = {at_angle_d}')

    for link in robot.links:
        for fe, cp in robot.indices(one_based=True):
            phi = link['q'][fe, cp, 'phi']
            phi.setub(radians(15))
            phi.setlb(radians(-15))

            psi = link['q'][fe, cp, 'psi']
            psi.setub(radians(10 + (at_angle_d or 0)))
            psi.setlb(radians(-10 + (at_angle_d or 0)))

    # bound theta
    for link in robot.links[:2]:  # body
        for fe, cp in robot.indices(one_based=True):
            link['q'][fe, cp, 'theta'].setub(radians(60))
            link['q'][fe, cp, 'theta'].setlb(radians(-60))

    for link in robot.links[2:]:  # everything else
        for fe, cp in robot.indices(one_based=True):
            link['q'][fe, cp, 'theta'].setub(radians(90))
            link['q'][fe, cp, 'theta'].setlb(radians(-90))

    # never fallen over
    for fe, cp in robot.indices(one_based=True):
        body['q'][fe, cp, 'z'].setlb(0.3)
        body['q'][fe, cp, 'z'].setub(0.7)

    if at_angle_d is None:
        # roughly bound to y plane
        for fe, cp in robot.indices(one_based=True, skipfirst=False):
            body['q'][fe, cp, 'y'].setub(0.2)
            body['q'][fe, cp, 'y'].setlb(-0.2)

        # average velocity init (overwrite the init!)
        for fe, cp in robot.indices(one_based=True, skipfirst=False):
            body['q'][fe, cp, 'x'].value = avg_vel * \
                total_time * (fe-1 + (cp-1)/ncp)/(nfe-1)
            body['dq'][fe, cp, 'x'].value = avg_vel

        body['q'][nfe, ncp, 'x'].fix(total_time*avg_vel)

        # periodic
        periodic(robot, but_not=('x',))
    else:
        θᵣ = radians(at_angle_d)

        # average velocity init (overwrite the init!)
        for fe, cp in robot.indices(one_based=True, skipfirst=False):
            scale = total_time * (fe-1 + (cp-1)/ncp)/(nfe-1)
            body['q'][fe, cp, 'x'].value = avg_vel * scale * math.cos(θᵣ)
            body['dq'][fe, cp, 'x'].value = avg_vel * math.cos(θᵣ)
            body['q'][fe, cp, 'y'].value = avg_vel * scale * math.sin(θᵣ)
            body['dq'][fe, cp, 'y'].value = avg_vel * math.sin(θᵣ)

        #ol.visual.warn('Should probably also bound x, y!')

        body['q'][nfe, ncp, 'x'].fix(total_time * avg_vel * math.cos(θᵣ))
        body['q'][nfe, ncp, 'y'].fix(total_time * avg_vel * math.sin(θᵣ))

        # periodic
        periodic(robot, but_not=('x', 'y'))

    return add_costs(robot, include_transport_cost=False, include_torque_cost=False)


# def set_quad_motor_limits(robot: System3D):
#     """
#     >>> robot.make_pyomo_model(nfe=10,  collocation='implicit_euler', total_time=0.3)
#     >>> increase_motor_limits(robot, torque_bound=5., no_load_speed=100.)
#     >>> ol.motor.torques(robot)[0]['Tc'].pprint()
#     """
#     assert robot.m is not None, \
#         'robot.make_pyomo_model() must be called before calling this function'

#     motors = {motor.name: motor for motor in ol.motor.torques(robot)}

#     def set_lims(name, torque_bound, no_load_speed):
#         motor = motors[name]
#         for Tc in motor_['Tc'][:, :]:
#             Tc.setub(+torque_bound)
#             Tc.setlb(-torque_bound)
#             if hasattr(motor, 'torque_speed_limit'):
#                 tsp = motor.torque_speed_limit
#                 tsp.torque_bounds = (-torque_bound, torque_bound)
#                 tsp.no_load_speed = no_load_speed

#     for name in ("base_B_base_F_torque", "base_B_UBL_torque", "base_B_UBR_torque"):
#         set_lims(name, 2.5, 75.)
#     for name in ("base_F_UFL_torque", "base_F_UFR_torque"):
#         set_lims(name, 2., 150.)
#     # for name in ("base_B_tail0_torque", "tail0_tail1_torque"):
#     #     set_lims(name, TORQUE, SPEED)
#     for name in ("UFL_LFL_torque", "UFR_LFR_torque"):
#         set_lims(name, 1., 75.)
#     for name in ("UBL_LBL_torque", "UBR_LBR_torque"):
#         set_lims(name, 0.75, 50.)
